import os
from subprocess import check_output


__author__     = "Sebastiano Ferraris UCL"
__licence__    = "MIT"
__repository__ = "https://github.com/SebastianoF/bruker2nifti"
__all__ = ['_cores', '_getters', '_utils', 'converter', 'open_GUI']

# here = os.path.abspath(os.path.dirname(__file__))
# git_dir = os.path.dirname(here)

# Describe the version relative to last tag
# command_git = ['git', 'describe', '--match', 'v[0-9]*']
# version_buf = check_output(command_git, cwd=git_dir).rstrip()

# Exclude the 'v' for PEP440 conformity, see
# https://www.python.org/dev/peps/pep-0440/#public-version-identifiers
__version__ = '1.0.3'  # version_buf[1:].decode("utf-8")



import os
import numpy as np
import warnings
import sys

from numpy.testing import assert_array_equal, assert_equal, assert_raises

from bruker2nifti._getters import get_stack_direction_from_VisuCorePosition


def test_get_stack_direction_from_VisuCorePosition_OK_dummy_multiple_cases():
    visu_core_position_ = np.array([[-20., -20.,  -4.],
                                    [-20., -20.,  -2.],
                                    [-20., -20.,   0.],
                                    [-20., -20.,   2.],
                                    [-20., -20.,   4.],
                                    [4., -20.,  20.],
                                    [2., -20.,  20.],
                                    [0., -20.,  20.],
                                    [-2., -20.,  20.],
                                    [-4., -20.,  20.],
                                    [-20.,  4., 20.],
                                    [-20.,  2., 20.],
                                    [-20., 0., 20.],
                                    [-20., - 2.,  20.],
                                    [-20., - 4., 20.]])
    a = get_stack_direction_from_VisuCorePosition(visu_core_position_, 3)
    assert_equal(a, 'z+x-y-')


def test_get_stack_direction_from_VisuCorePosition_OK_dummy_single_cases():
    vcp1 = np.array([[-4., -20.,  20.], [-2., -20., 20.],  [0., -20., 20.], [2., -20.,  20.], [4., -20., 20.]])
    vcp2 = np.array([[4., -20.,  20.], [2., -20.,  20.],  [0., -20.,  20.], [-2., -20.,  20.],  [-4., -20., 20.]])
    vcp3 = np.array([[-20.,  -4.,  20.], [-20.,  -2.,  20.], [-20.,   0.,  20.], [-20.,  2.,  20.], [-20.,  4., 20.]])
    vcp4 = np.array([[-20.,   4.,  20.], [-20.,   2.,  20.], [-20.,   0.,  20.], [-20.,  -2.,  20.], [-20.,  -4., 20.]])
    vcp5 = np.array([[-20., -20.,  -4.], [-20., -20.,  -2.], [-20., -20.,   0.], [-20., -20.,   2.], [-20., -20.,  4.]])
    vcp6 = np.array([[-20., -20.,  4.], [-20., -20.,  2.],  [-20., -20.,   0.], [-20., -20., -2.], [-20., -20., -4.]])

    a1 = get_stack_direction_from_VisuCorePosition(vcp1, 1)
    a2 = get_stack_direction_from_VisuCorePosition(vcp2, 1)
    a3 = get_stack_direction_from_VisuCorePosition(vcp3, 1)
    a4 = get_stack_direction_from_VisuCorePosition(vcp4, 1)
    a5 = get_stack_direction_from_VisuCorePosition(vcp5, 1)
    a6 = get_stack_direction_from_VisuCorePosition(vcp6, 1)

    assert_equal(a1, 'x+')
    assert_equal(a2, 'x-')
    assert_equal(a3, 'y+')
    assert_equal(a4, 'y-')
    assert_equal(a5, 'z+')
    assert_equal(a6, 'z-')


def test_get_stack_direction_from_VisuCorePosition_wrong_shape1():
    vcp = np.array([[-4., -20., 20.]])
    with assert_raises(IOError):
        get_stack_direction_from_VisuCorePosition(vcp, 1)
    vcp = np.array([-4., -20., 20.])
    with assert_raises(IOError):
        get_stack_direction_from_VisuCorePosition(vcp, 1)


def test_get_stack_direction_from_VisuCorePosition_wrong_shape2():
    vcp = np.array([[[-4., -20., 20.], [-2., -20., 20.]], [[0., -20., 20.], [2., -20., 20.]]])
    with assert_raises(IOError):
        get_stack_direction_from_VisuCorePosition(vcp, 1)


def test_get_stack_direction_from_VisuCorePosition_wrong_shape3():
    vcp = np.array([[-20., -20.], [-20., -20.],  [-20., -20.], [-20., -20.], [-20., -20.]])
    with assert_raises(IOError):
        get_stack_direction_from_VisuCorePosition(vcp, 1)


def test_get_stack_direction_from_VisuCorePosition_wrong_sub_volumes():
    visu_core_position_ = np.array([[-20., -20., -4.],
                                    [-20., -20., -2.],
                                    [-20., -20., 0.],
                                    [-20., -20., 2.],
                                    [-20., -20., 4.],
                                    [4., -20., 20.],
                                    [2., -20., 20.],
                                    [0., -20., 20.],
                                    [-2., -20., 20.],
                                    [-4., -20., 20.],
                                    [-20., 4., 20.],
                                    [-20., 2., 20.],
                                    [-20., 0., 20.],
                                    [-20., - 2., 20.],
                                    [-20., - 4., 20.]])
    with assert_raises(IOError):
        get_stack_direction_from_VisuCorePosition(visu_core_position_, 2)



import numpy as np
import os
import nibabel as nib
import warnings
from os.path import join as jph


# --- text-files utils ---

def unique_words_in_string(in_string):
    ulist = []
    [ulist.append(s) for s in in_string.split() if s not in ulist]
    return ulist[0]


def indians_file_parser(s, sh=None):
    """
    An here-called indians file is a string obtained from a sequence of rows from a Bruker parameter file
    whose shape needs to be changed, in function of its content and according to an optional parameter sh
    that defines the shape of the output.
    This function transform the indian file in a data structure,
    according to the information that can be parsed in the file:
    A - list of vectors transformed into a list
    B - list of numbers, transformed into a np.ndarray, or single number stored as a float.
    B bis - string of 'inf' repeated n times that will be transformed in a numpy array of 'inf'.
    C - list of strings separated by <>.
    D - everything else becomes a string.

    :param s: string indian file
    :param sh: shape related
    :return: parsed indian file of adequate output.
    """

    s = s.strip()  # removes initial and final spaces.

    # A
    if ('(' in s) and (')' in s):
        s = s[1:-1]  # removes initial and final ( )
        a = ['(' + v + ')' for v in s.split(') (')]
    # B
    elif s.replace('-', '').replace('.', '').replace(' ', '').replace('e', '').isdigit():
        if ' ' in s:
            a = np.array([float(x) for x in s.split()])
            if sh is not None:
                a = a.reshape(sh)
        else:
            a = float(s)
    # B-bis
    elif 'inf' in s:
        if 'inf' == unique_words_in_string(s):
            num_occurrences = sum('inf' == word for word in s.split())
            a = [np.inf, ] * num_occurrences
        else:
            a = s[:]
    # C
    elif ('<' in s) and ('>' in s):
        s = s[1:-1]  # removes initial and final < >
        a = [v for v in s.split('> <')]
    # D
    else:
        a = s[:]

    # added to work with ParaVision vers 6.0.1:
    if isinstance(a, list):
        if len(a) == 1:
            a = a[0]

    return a


def var_name_clean(line_in):
    """
    Removes #, $ and PVM_ from line_in, where line in is a string from a Bruker parameter list file.
    :param line_in: input string
    :return: output string cleaned from #, $ and PVM_
    """
    line_out = line_in.replace('#', '').replace('$', '').replace('PVM_', '').strip()
    return line_out


def from_dict_to_txt_sorted(dict_input, pfi_output):
    """
    Simple auxiliary to save the information contained in a dictionary into a txt file
    at the specified path to file (pfi).
    :param dict_input: input structure dictionary
    :param pfi_output: path to file.
    :return:
    """
    sorted_keys = sorted(dict_input.keys())

    with open(pfi_output, 'w') as f:
        f.writelines('{0} = {1} \n'.format(k, dict_input[k]) for k in sorted_keys)


def bruker_read_files(param_file, data_path, sub_scan_num='1'):
    """
    Reads parameters files of from Bruker raw data imaging format.
    It parses the files 'acqp', 'method', 'reco', 'visu_pars' and 'subject'.
    Even if only 'visu_pars' is relevant for the conversion to nifti, having a more general parser has turned out
    to be useful in many cases (e.g. in PV5.1 to check).
    :param param_file: file parameter, must be a string in the list ['acqp', 'method', 'reco', 'visu_pars', 'subject'].
    :param data_path: path to data.
    :param sub_scan_num: number of the sub-scan folder where usually the 'reco' and 'visu_pars' parameter files
    are stored.
    :return: dict_info dictionary with the parsed information from the input file.
    """
    if param_file.lower() == 'reco':
        if os.path.exists(jph(data_path, 'pdata', '1', 'reco')):
            f = open(jph(data_path, 'pdata', '1', 'reco'), 'r')
        else:
            print('File {} does not exist'.format(jph(data_path, 'pdata', '1', 'reco')))
            return {}
    elif param_file.lower() == 'acqp':
        if os.path.exists(jph(data_path, 'acqp')):
            f = open(jph(data_path, 'acqp'), 'r')
        else:
            print('File {} does not exist'.format(jph(data_path, 'acqp')))
            return {}
    elif param_file.lower() == 'method':
        if os.path.exists(jph(data_path, 'method')):
            f = open(jph(data_path, 'method'), 'r')
        else:
            print('File {} does not exist'.format(jph(data_path, 'method')))
            return {}
    elif param_file.lower() == 'visu_pars':
        if os.path.exists(jph(data_path, 'pdata', str(sub_scan_num), 'visu_pars')):
            f = open(jph(data_path, 'pdata', str(sub_scan_num), 'visu_pars'), 'r')
        elif os.path.exists(jph(data_path, str(sub_scan_num), 'pdata', '1', 'visu_pars')):
            f = open(jph(data_path, str(sub_scan_num), 'pdata', '1', 'visu_pars'), 'r')
        else:
            print('File {} does not exist'.format(jph(data_path, 'pdata', str(sub_scan_num), 'visu_pars')))
            return {}
    elif param_file.lower() == 'subject':
        if os.path.exists(jph(data_path, 'subject')):
            f = open(jph(data_path, 'subject'), 'r')
        else:
            print('File {} does not exist'.format(jph(data_path, 'subject')))
            return {}
    else:
        raise IOError("param_file input must be the string 'reco', 'acqp', 'method', 'visu_pars' or 'subject'")

    dict_info = {}
    lines = f.readlines()

    for line_num in range(len(lines)):
        '''
        Relevant information are in the lines with '##'.
        For the parameters that have arrays values specified between (), with values in the next line.
        Values in the next line can be parsed in lists or np.ndarray when they contains also characters or numbers.
        '''

        line_in = lines[line_num]

        if '##' in line_in:

            if ('$' in line_in) and ('(' in line_in) and ('<' not in line_in):
                # A:
                splitted_line = line_in.split('=')
                # name of the variable contained in the row, and shape:
                var_name = var_name_clean(splitted_line[0][3:])

                done = False
                indian_file = ''
                pos = line_num
                sh = splitted_line[1]
                # this is not the shape of the vector but the beginning of a full vector.
                if sh.replace(' ', '').endswith(',\n'):
                    sh = sh.replace('(', '').replace(')', '').replace('\n', '').strip()
                    indian_file += sh
                    sh = None
                # this is not the shape of the vector but a full vector.
                elif sh.replace(' ', '').endswith(')\n') and '.' in sh:
                    sh = sh.replace('(', '').replace(')', '').replace('\n', '').strip()
                    indian_file += sh
                    sh = None
                # this is finally the shape of the vector that will start in the next line.
                else:
                    sh = sh.replace('(', '').replace(')', '').replace('\n', '').strip()
                    sh = [int(num) for num in sh.split(',')]

                while not done:

                    pos += 1
                    # collect the indian file: info related to the same variables that can appears on multiple rows.
                    line_to_explore = lines[pos]  # tell seek does not work in the line iterators...

                    if ('##' in line_to_explore) or ('$$' in line_to_explore):
                        # indian file is over
                        done = True

                    else:
                        # we store the rows in the indian file all in the same string.
                        indian_file += line_to_explore.replace('\n', '').strip() + ' '

                dict_info[var_name] = indians_file_parser(indian_file, sh)

            elif ('$' in line_in) and ('(' not in line_in):
                # B:
                splitted_line = line_in.split('=')
                var_name = var_name_clean(splitted_line[0][3:])
                indian_file = splitted_line[1]

                dict_info[var_name] = indians_file_parser(indian_file)

            elif ('$' not in line_in) and ('(' in line_in):
                # C:
                splitted_line = line_in.split('=')
                var_name = var_name_clean(splitted_line[0][2:])

                done = False
                indian_file = splitted_line[1].strip() + ' '
                pos = line_num

                while not done:
                    pos += 1
                    # collect the indian file: info related to the same variables that can appears on multiple rows.
                    line_to_explore = lines[pos]  # tell seek does not work in the line iterators...
                    if ('##' in line_to_explore) or ('$$' in line_to_explore):
                        # indian file is over
                        done = True
                    else:
                        # we store the rows in the indian file all in the same string.
                        indian_file += line_to_explore.replace('\n', '').strip() + ' '

                dict_info[var_name] = indians_file_parser(indian_file)

            elif ('$' not in line_in) and ('(' not in line_in):
                # D:
                splitted_line = line_in.split('=')
                var_name = var_name_clean(splitted_line[0])
                indian_file = splitted_line[1].replace('=', '').strip()
                dict_info[var_name] = indians_file_parser(indian_file)

            else:
                # General case: take it as a simple string.
                splitted_line = line_in.split('=')
                var_name = var_name_clean(splitted_line[0])
                dict_info[var_name] = splitted_line[1].replace('(', '').replace(')', '').replace('\n', ''). \
                                                       replace('<', '').replace('>', '').replace(',', ' ').strip()

        else:
            # line does not contain any 'assignable' variable, so this information is not included in the info.
            pass

    return dict_info


# --- Slope correction utils ---


def eliminate_consecutive_duplicates(input_list):
    """
    Simple funcion to eliminate consecutive duplicates in a list or arrays or in a list of numbers.
    :param input_list: list with possible consecutive duplicates.
    :return: input_list with no consecutive duplicates.
    """
    if isinstance(input_list[0], np.ndarray):
        output_list = [input_list[0]]
        for k in input_list[1:]:
            if not list(k) == list(output_list[-1]):
                output_list.append(k)
        return output_list
    else:
        output_list = [input_list[0], ]
        for i in range(1, len(input_list)):
            if not input_list[i] == input_list[i-1]:
                output_list.append(input_list[i])
        return output_list


def data_corrector(data, factors, kind='slope', num_initial_dir_to_skip=None, dtype=np.float64):
    """
    Slope is a float or a vector that needs to be multiplied to the data, to obtain the data as they are acquired.
    To reduce the weight of an image, each slice can be divided by a common float factor, so that at each voxel only the
    integer remaining is stored:

    real_value_acquired[slice_j][x] = data_integer_reminder[slice_j][x] * float_slope[slice_j][x]

    (where = is an almost equal, where the small loss of accuracy is justified by the huge amount of space saved)

    :param data: data as parsed from the data structure.
    :param factors: can be the slope or the offset as parsed from the data structure
    :param kind: is a string that can be 'slope' (multiplicative factor) or 'offset' additive factor.
    :param num_initial_dir_to_skip: in some cases (as some DWI) the number of slices in the image is higher than the
    provided slope/offset length. Usually it is because the initial directions have no weighted and the first element
    in the slope/offset can correct them all. If num_initial_direction_to_skip=j the slope/offset correction starts
    after j slices, and the initial j timepoint are trimmed by j.
    :param dtype: [np.float64] output datatype.
    :return: data after the slope/offset correction.
    ---
    NOTE 1: if used in sequence to correct for slope and offset, correct FIRST slope, then OFFSET.
    NOTE 2: when read 'factor' think slope or offset. The two are embeded in the same method to avoid code repetition.
    """

    if len(data.shape) > 5:
        raise IOError('4d or lower dimensional images allowed. Input data has shape {} '.format(data.shape))
    assert kind in ('slope', 'offset')

    if hasattr(factors, '__contains__'):
        if np.inf in factors:
            warnings.warn('bruker2nifti - Vector corresponding to {} has some inf values. Can not correct it.'.
                          format(kind), UserWarning)
            return data

    data = data.astype(dtype)

    if num_initial_dir_to_skip is not None:
        factors = factors[num_initial_dir_to_skip:]
        data = data[..., num_initial_dir_to_skip:]

    # Check compatibility slope and data and if necessarily correct for possible consecutive duplicates
    # (as in some cases, when the size of the slope is larger than any timepoint or spatial point, the problem can
    # be in the fact that there are duplicates in the slope vector. This has been seein only in PV5.1).
    if not (isinstance(factors, int) or isinstance(factors, float)):
        if factors.ndim == 1:
            if not factors.size == data.shape[-1] and not factors.size == data.shape[-2]:
                factors = np.array(eliminate_consecutive_duplicates(list(factors)), dtype=np.float64)
                if not factors.size == data.shape[-1] and not factors.size == data.shape[-2]:
                    msg = 'Slope shape {0} and data shape {1} appears to be not compatible'.format(
                        factors.shape, data.shape)
                    raise IOError(msg)

    if isinstance(factors, int) or isinstance(factors, float):
        # scalar slope/offset times nd array data
        if kind == 'slope':
            data *= factors
        elif kind == 'offset':
            data += factors

    elif factors.size == 1:
        # scalar slope/offset embedded in a singleton times nd array data
        if kind == 'slope':
            data *= factors[0]
        else:
            data += factors[0]

    elif len(data.shape) == 3 and len(factors.shape) == 1:
        # each slice of the 3d image is multiplied an element of the slope consecutively
        if data.shape[2] == factors.shape[0]:
            for t, fa in enumerate(factors):
                if kind == 'slope':
                    data[..., t] = data[..., t] * fa
                elif kind == 'offset':
                    data[..., t] = data[..., t] + fa
        else:
            raise IOError('Shape of the 2d image and slope dimensions are not consistent')

    elif len(data.shape) == 4 and len(factors.shape) == 1 and factors.shape[0] == data.shape[2]:
        # each slice of the 4d image, taken from the third dim, is multiplied by each element of the slope in sequence.
        if factors.size == data.shape[2]:
            for t in range(data.shape[3]):
                for k in range(factors.size):
                    if kind == 'slope':
                        data[..., k, t] = data[..., k, t] * factors[k]
                    elif kind == 'offset':
                        data[..., k, t] = data[..., k, t] + factors[k]
        else:
            raise IOError('If you are here, your case cannot be converted. Further investigations required.')

    elif len(data.shape) == 5 and len(factors.shape) == 1 and factors.shape[0] == data.shape[3]:
        # each slice of the 5d image, taken from the fourth dim, is multiplied by each element of the slope in sequence.
        if factors.size == data.shape[3]:
            for t in range(data.shape[4]):
                for k in range(factors.size):
                    if kind == 'slope':
                        data[..., k, t] = data[..., k, t] * factors[k]
                    elif kind == 'offset':
                        data[..., k, t] = data[..., k, t] + factors[k]
        else:
            raise IOError('If you are here, your case cannot be converted. Further investigations required.')

    else:
        # each slice of the nd image, taken from the last dimension, is multiplied by each element of the slope.
        if factors.size == data.shape[-1]:
            for t in range(data.shape[-1]):
                if kind == 'slope':
                    data[..., t] = data[..., t] * factors[t]
                elif kind == 'offset':
                    data[..., t] = data[..., t] + factors[t]
        else:
            msg = 'Slope shape {0} and data shape {1} appears to be not compatible'.format(factors.shape, data.shape)
            raise IOError(msg)

    return data


# -- nifti affine matrix utils --


def compute_resolution_from_visu_pars(vc_extent, vc_size, vc_frame_thickness):
    """
    Resolution parameter is provided as a vector in the 'reco' parameter file. To extract the information from the
    'visu_pars' only, as some scans can lack the reco file, some computation on its paramteres neesd to be performed.
    :param vc_extent: VisuCoreExtent parameter file from 'visu_pars'.
    :param vc_size: VisuCoreSize parameter file from 'visu_pars'.
    :param vc_frame_thickness: VisuCoreFrameThickness parameter file from 'visu_pars'.
    :return:
    """

    if len(vc_extent) == len(vc_size):
        resolution = [e / float(s) for e, s in zip(vc_extent, vc_size)]
    else:
        raise IOError

    if isinstance(vc_frame_thickness, np.ndarray) or isinstance(vc_frame_thickness, list):
        vc_frame_thickness = vc_frame_thickness[0]

    if len(vc_extent) == 2:
        resolution += [vc_frame_thickness]
        return resolution
    elif len(vc_extent) == 3:
        return resolution
    else:
        raise IOError


def sanity_check_visu_core_subject_position(vc_subject_position):
    """
    The parameter VisuCoreSubjectPosition can be 'Head_Prone' or 'Head_Supine'. Tertium non datur.
    :param vc_subject_position: VisuCoreSubjectPosition from 'visu_pars'
    :return: Raise error if VisuCoreSubjectPosition is not 'Head_Prone' or 'Head_Supine'
    """
    if vc_subject_position not in ['Head_Prone', 'Head_Supine']:
        msg = "Known cases are 'Head_Prone' or  'Head_Supine' for the parameter 'visu_pars.VisuSubjectPosition."
        raise IOError(msg)


def filter_orientation(visu_parse_orientation):
    """
    Pre-process the paramter value VisuParseOrientation from the 'visu_pars' paramter file.
    :param visu_parse_orientation: VisuParseOrientation from the 'visu_pars' paramter file.
    :return: re-shaped and rounded VisuParseOrientation parameter.
    """

    if not np.prod(visu_parse_orientation.shape) == 9:
        # Take the first 9 elements:
        visu_parse_orientation = visu_parse_orientation.flat[:9]

    ans = np.around(visu_parse_orientation.reshape([3, 3], order='F'), decimals=4)
    return ans


def pivot(v):
    """
    :param v: vector or list
    :return: max in absolute value with original sign or max from origin.
    Corresponds to the main direction for each column of an orientation matrix.
    """
    return v[list(abs(v)).index(abs(v).max())]


def compute_affine_from_visu_pars(vc_orientation, vc_position, vc_subject_position, resolution,
                                  frame_body_as_frame_head=False, keep_same_det=True,
                                  consider_subject_position=False):
    """
    How the affine is computed (to the understanding acquired so far):

    0) resolution, orientation and translation are provided in separate arrays, we combine them together in a
    standard 4x4 matrix.

    1) We invert the resulting matrix - according to conventions ParaVision (scanner to image frame)
       and DICOM/Nifti (image to scanner frame).

    2) impose the signs of the first two columns (pivots) to be negative, and the third to be be positive.
    - according to the fact that the provided transformation is DICOM-like (LPS) instead of NIFTI like (RAS)
    (Left/Right, Anterior/Posterior, Inferior/Superior).

    -------- optional changes ----------

    3) frame_body_as_frame_head: Switching the last 2 columns of the rotational part, no matter the value of
    VisuCorePosition - According to the fact we are dealing with quadrupeds and not with humans,
    we need to switch the Anterior-Posterior with the Inferior-Superior direction.
    Set frame_body_as_frame_head=True to set the biped orientation.

    4) consider_subject_position: This can be 'head_prone' or 'head_supine'.
    Reason why sometimes this must be considered for a correct
    orientation and must be considered dis-jointly with frame_body_as_frame_head, is that this parameter is sometimes
    tuned to voluntarily switch from radiological to neurological coordinate systems.
    If the subject is Prone and the technician wants to have the coordinates in neurological he/she can consciously
    set the variable vc_subject_position to 'Head_Supine', even if the subject is not supine.

    5) keep_same_det: Finally, for safety, we can impose the same determinant as the input matrix.

    (If there is any b-vectors list, this is modified accordingly).

    :param vc_orientation: visu core orientation parameter.
    :param vc_position: visu core position parameter. -  corresponds to the translational part of the matrix.
    :param vc_subject_position: 'Head_Prone' or 'Head_Supine'. If head supine and if consider_subject_position is True
    it invert the direction of the axis anterior-posterior. - do not confuse subject_position with positon (read this
    last as 'translation').
    :param resolution: resolution of the image, output of compute_resolution_from_visu_pars in the same module.
    :param frame_body_as_frame_head: [False] to parametrise the difference between monkeys [True] and rats [False].
    :param keep_same_det: in case you want the determinant to be the same as the input one. Consider it in particular
    if frame_body_as_frame_head is set to False, and according to the choice of consider_subject_position.
    :param consider_subject_position: [False] The reason why sometimes this must be considered for a correct
    orientation and sometimes must not, is that this parameter is tuned to voluntarily switch from radiological
    to neurological coordinate systems. If the subject is Prone and the technician wants to have the coordinates
    in neurological he/she can consciously set the variable vc_subject_position to 'Head_Supine'.
    :return: final affine (qform) transformation according to the nifti convention

    NOTE: we are assuming that the angles parametrisation is the same for the input and the output.
    We hope this is the case as we do not have any mean to confirm that. The fslreorient2std from FSL
    should be applied afterwards to all the images (after DWI analysis if any).
    """

    sanity_check_visu_core_subject_position(vc_subject_position)
    vc_orientation = filter_orientation(vc_orientation)

    # 0) integrate resolution with the orientation and add the translation in the projective coordinates:

    result = np.eye(4, dtype=np.float32)
    result[0:3, 0:3] = vc_orientation
    result[0:3, 3] = vc_position

    # 1) Invert the orientation matrix, according to nifti convention and Bruker manual.
    # Round the decimals to avoid precision problems. Check if determinant makes sense.
    result = np.round(np.linalg.inv(result), decimals=4)
    result_det = np.linalg.det(result)
    if result_det == 0:
        raise IOError('Orientation determinant is 0. Cannot grasp this dataset.')

    # 2-3) impose pivot first column negative, second column negative, third column positive
    result_orientation = result[:3, :3]

    result_orientation = result_orientation.dot(np.array([[1, 0, 0], [0, 0, 1], [0, 1, 0]]))
    if frame_body_as_frame_head:  # from SAR to ASL
        result_orientation = result_orientation.dot(np.array([[0, -1, 0], [1, 0, 0], [0, 0, 1]]))

    if pivot(result_orientation[:, 0]) > 0:
        result_orientation[:, 0] = -1 * result_orientation[:, 0]
    if pivot(result_orientation[:, 1]) > 0:
        result_orientation[:, 1] = -1 * result_orientation[:, 1]
    if pivot(result_orientation[:, 2]) < 0:
        result_orientation[:, 2] = -1 * result_orientation[:, 2]

    result_orientation = result_orientation.dot(np.diag(resolution))

    result[:3, :3] = result_orientation

    # 4) - optional
    if consider_subject_position:
        if vc_subject_position == 'Head_Prone':
            result[1, :] = -1 * result[1, :]
    # 5) - optional
    if keep_same_det:
        if (np.linalg.det(result) < 0 < result_det) or (np.linalg.det(result) > 0 > result_det):
            result[0, :3] = -1 * result[0, :3]

    return result


# --- b-vectors utils ---


def obtain_b_vectors_orient_matrix(vc_orientation, vc_subject_position, frame_body_as_frame_head=False,
                                           keep_same_det=True, consider_subject_position=False):

    """
    See _utils.compute_affine_from_visu_pars help for the same input parameters.
    :param vc_orientation: VisuCoreOrientation parameter file
    :param vc_subject_position: VisuCoreSubjectPosition parameter file
    :param frame_body_as_frame_head:
    :param keep_same_det:
    :param consider_subject_position:
    :return:
    """
    resolution = np.array([1, 1, 1])
    translation = np.array([0, 0, 0])

    aff = compute_affine_from_visu_pars(vc_orientation, translation, vc_subject_position, resolution,
                                        frame_body_as_frame_head=frame_body_as_frame_head, keep_same_det=keep_same_det,
                                        consider_subject_position=consider_subject_position)

    return np.copy(aff[:3, :3])


def normalise_b_vect(b_vect, remove_nan=True):
    """
    Normalisation of the b_vector matrix (dim : num b-vectors x 3)
    :param b_vect: the b_vector matrix (dim : num b-vectors x 3)
    :param remove_nan: remove nan if appears in the b-vector matrix, applying np.nan_to_num.
    :return: normalised b-vectors.
    """

    b_vect_normalised = np.zeros_like(b_vect)
    norms = np.linalg.norm(b_vect, axis=1)

    for r in range(b_vect.shape[0]):
        if norms[r] < 10e-5:
            b_vect_normalised[r, :] = np.nan
        else:
            b_vect_normalised[r, :] = (1 / float(norms[r])) * b_vect[r, :]

    if remove_nan:
        b_vect_normalised = np.nan_to_num(b_vect_normalised)

    return b_vect_normalised


def apply_reorientation_to_b_vects(reorientation_matrix, row_b_vectors_in_rows):
    """
    :param reorientation_matrix: a 3x3 matrix representing a reorientation in the 3D space:
    Typically with det = 1 or -1.
    a b c
    d e f
    g h i

    :param row_b_vectors_in_rows:
     A nx3 matrix where n row-major b-vectors (v1, v2, v3, v4, ...) are aligned in rows
    v1_1 v1_2 v1_3
    v2_1 v2_2 v2_3
    v3_1 v3_2 v3_3
    v4_1 v4_2 v4_3
    ...

    :return:
    An nx3 matrix where each row is the corresponding b-vector multiplied by the same matrix reorientation_matrix:
    a.v1_1 +  b.v1_2 + c.v1_3 + d.v1_1 +  e.v1_2 + f.v1_3 + g.v1_1 +  h.v1_2 + i.v1_3
    a.v2_1 +  b.v2_2 + c.v2_3 + d.v2_1 +  e.v2_2 + f.v2_3 + g.v2_1 +  h.v2_2 + i.v2_3
    a.v3_1 +  b.v3_2 + c.v3_3 + d.v3_1 +  e.v3_2 + f.v3_3 + g.v3_1 +  h.v3_2 + i.v3_3
    a.v4_1 +  b.v4_2 + c.v4_3 + d.v4_1 +  e.v4_2 + f.v4_3 + g.v4_1 +  h.v4_2 + i.v4_3
    ...

    """
    b_vectors_in_column_reoriented = np.einsum('ij, kj -> ki', reorientation_matrix, row_b_vectors_in_rows)
    return b_vectors_in_column_reoriented


# -- nibabel-related utils --


def set_new_data(image, new_data, new_dtype=None, remove_nan=True):
    """
    From a nibabel image and a numpy array it creates a new image with
    the same header of the image and the new_data as its data.
    :param image: nibabel image
    :param new_data: numpy array
    :param new_dtype:
    :param remove_nan:
    :return: nibabel image
    """
    if remove_nan:
        new_data = np.nan_to_num(new_data)

    # if nifty1
    if image.header['sizeof_hdr'] == 348:
        new_image = nib.Nifti1Image(new_data, image.affine, header=image.header)
    # if nifty2
    elif image.header['sizeof_hdr'] == 540:
        new_image = nib.Nifti2Image(new_data, image.affine, header=image.header)
    else:
        raise IOError('Input image header problem')

    # update data type:
    if new_dtype is None:
        new_image.set_data_dtype(new_data.dtype)
    else:
        new_image.set_data_dtype(new_dtype)

    return new_image



import numpy as np
import nibabel as nib

from numpy.testing import assert_array_equal, assert_equal, assert_raises, assert_almost_equal

from bruker2nifti._utils import indians_file_parser, normalise_b_vect, data_corrector, \
    eliminate_consecutive_duplicates, compute_resolution_from_visu_pars, compute_affine_from_visu_pars, \
    apply_reorientation_to_b_vects, set_new_data, obtain_b_vectors_orient_matrix


# --- TEST text-files utils ---


def test_indians_file_parser_A():

    indian_file_test_1 = "('<VisuCoreOrientation>, 0') ('<VisuCorePosition>, 0')"
    indian_file_test_2 = "(EPI (pmv))"
    indian_file_test_3 = "(5, <FG_SLICE>, <>, 0, 2)"
    indian_file_test_4 = "( SPAM )"

    a1 = indians_file_parser(indian_file_test_1)
    a2 = indians_file_parser(indian_file_test_2)
    a3 = indians_file_parser(indian_file_test_3)
    a4 = indians_file_parser(indian_file_test_4)

    assert_equal(a1, ["('<VisuCoreOrientation>, 0')", "('<VisuCorePosition>, 0')"])
    assert_equal(a2, indian_file_test_2)
    assert_equal(a3, indian_file_test_3)
    assert_equal(a4, indian_file_test_4)


def test_indians_file_parser_B():

    indian_file_test_1 = "4.24"
    indian_file_test_2 = "  4.24  "
    indian_file_test_3 = "1"
    indian_file_test_4 = "80 64e2"
    indian_file_test_5 = "22.3 5.3e12 4.4 5.62e-5 3.9 42.42"

    a1 = indians_file_parser(indian_file_test_1)
    a2 = indians_file_parser(indian_file_test_2)
    a3 = indians_file_parser(indian_file_test_3)
    a4 = indians_file_parser(indian_file_test_4, sh=[2])
    a5 = indians_file_parser(indian_file_test_5, sh=[2, 3])
    a6 = indians_file_parser(indian_file_test_5)

    assert_equal(a1, 4.24)
    assert_equal(a2, 4.24)
    assert_equal(a3, 1)
    np.testing.assert_array_equal(a4, np.array([80, 6400]))
    np.testing.assert_array_equal(a5, np.array([[22.3, 5.3e12, 4.4], [5.62e-05, 3.9, 42.42]]))
    np.testing.assert_array_equal(a6, np.array([22.3, 5.3e12, 4.4, 5.62e-05, 3.9, 42.42]))


def test_indians_file_parser_C():

    indian_file_test_1 = "<123.321.123>"
    indian_file_test_2 = "<123.321.123> <123.321.123>"
    indian_file_test_3 = "<20:19:47  1 Sep 2016>"

    a1 = indians_file_parser(indian_file_test_1)
    a2 = indians_file_parser(indian_file_test_2)
    a3 = indians_file_parser(indian_file_test_3)

    assert_equal(a1, '123.321.123')
    assert_equal(a2, ['123.321.123', '123.321.123'])
    assert_equal(a3, '20:19:47  1 Sep 2016')


def test_indians_file_parser_D():

    indian_file_test_1 = "Ultra-Spam"
    indian_file_test_2 = "OtherAnimal"
    indian_file_test_3 = "Prone_Supine"

    a1 = indians_file_parser(indian_file_test_1)
    a2 = indians_file_parser(indian_file_test_2)
    a3 = indians_file_parser(indian_file_test_3)

    assert_equal(a1, indian_file_test_1)
    assert_equal(a2, indian_file_test_2)
    assert_equal(a3, indian_file_test_3)


# --- TEST slope correction utils ---


def test_slope_corrector_int_float_slope():

    in_data = np.random.normal(5, 10, [3, 4, 5])
    sl1 = 5
    sl2 = 5.3

    out_data1 = data_corrector(in_data, sl1, kind='slope', dtype=np.float64)
    out_data2 = data_corrector(in_data, sl2, kind='slope', dtype=np.float64)

    assert_equal(out_data1.dtype, np.float64)
    assert_equal(out_data2.dtype, np.float64)
    assert_array_equal(out_data1, sl1 * in_data)
    assert_array_equal(out_data2, sl2 * in_data)


def test_slope_corrector_slice_wise_slope_3d():

    in_data = np.random.normal(5, 10, [4, 5, 3])
    sl = np.random.normal(5, 10, 3)

    out_data = data_corrector(in_data, sl, kind='slope', dtype=np.float64)

    for k in range(3):
        assert_array_equal(out_data[..., k], in_data[..., k]*sl[k])


def test_slope_corrector_slice_wise_slope_3d_fail():

    in_data = np.random.normal(5, 10, [4, 5, 3])
    sl = np.random.normal(5, 10, 5)

    with assert_raises(IOError):
        data_corrector(in_data, sl)


def test_slope_corrector_slice_wise_slope_4d():

    in_data = np.random.normal(5, 10, [2, 3, 4, 5])
    sl = np.random.normal(5, 10, 4)
    out_data = data_corrector(in_data, sl, kind='slope', dtype=np.float64)

    for t in range(5):
        for k in range(4):
            assert_array_equal(out_data[..., k, t], in_data[..., k, t] * sl[k])

# def test_slope_corrector_slice_wise_slope_4d_fail():
#
#     in_data = np.random.normal(5, 10, [2, 3, 4, 5])
#     sl = np.random.normal(5, 10, 5)
#
#     with assert_raises(IOError):
#         slope_corrector(in_data, sl)


def test_slope_corrector_slice_wise_slope_5d():

    in_data = np.random.normal(5, 10, [2, 3, 4, 5, 6])
    sl = np.random.normal(5, 10, 5)
    out_data = data_corrector(in_data, sl, kind='slope', dtype=np.float64)

    for t in range(6):
        for k in range(5):
            assert_array_equal(out_data[..., k, t], in_data[..., k, t] * sl[k])


def test_slope_corrector_slice_wise_slope_5d_fail():

    in_data = np.random.normal(5, 10, [2, 3, 4, 5, 6])
    sl = np.random.normal(5, 10, 7)

    with assert_raises(IOError):
        data_corrector(in_data, sl, kind='slope')

test_slope_corrector_slice_wise_slope_5d_fail()


# -- TEST nifti affine matrix utils --


def test_compute_resolution_from_visu_pars():

    # 1
    vc_extent = [25.0, 25.0]
    vc_size = [5, 5]
    vc_frame_thickness = 0.123
    res = compute_resolution_from_visu_pars(vc_extent, vc_size, vc_frame_thickness)

    assert_array_equal(res, [5, 5, 0.123])

    # 2
    vc_extent = [30.0, 30.0]
    vc_size = [5, 5]
    vc_frame_thickness = [0.123, 0.124, 0.125]
    res = compute_resolution_from_visu_pars(vc_extent, vc_size, vc_frame_thickness)

    assert_array_equal(res, [6.0, 6.0, 0.123])


# --- TEST b-vectors utils ---


def test_normalise_b_vect():

    num_vects = 30
    v = np.random.normal(5, 10, [num_vects, 3])
    v[5, :] = np.array([np.nan, ] * 3)
    v_normalised = normalise_b_vect(v)

    for k in list(set(range(num_vects)) - {5}):
        assert_almost_equal(np.linalg.norm(v_normalised[k, :]), 1.0)

    assert_equal(np.linalg.norm(v_normalised[5, :]), .0)


def test_apply_reorientation_to_b_vects_1():

    v = np.array([[1, 2, 3],
                  [4, 5, 6],
                  [7, 8, 9],
                  [10, 11, 12]])

    m = np.array([[1, 0, 0],
                  [0, 0, 1],
                  [0, 1, 0]])

    u = np.array([[1, 3, 2],
                  [4, 6, 5],
                  [7, 9, 8],
                  [10, 12, 11]])

    w = apply_reorientation_to_b_vects(m, v)

    np.testing.assert_array_equal(u, w)


def test_apply_reorientation_to_b_vects_2():

    v = np.array([[1,  2,  3],
                  [4,  5,  6],
                  [7,  8,  9],
                  [10, 11, 12]])

    m = np.array([[0, 1, 0],
                  [0, 0, 1],
                  [1, 0, 0]])

    u = np.array([[2,  3, 1],
                  [5,  6, 4],
                  [8,  9, 7],
                  [11, 12, 10]])

    w = apply_reorientation_to_b_vects(m, v)

    np.testing.assert_array_equal(u, w)


# -- TEST housekeeping utils --


def test_eliminate_consecutive_duplicates():

    l = [[3, 1, 4, 1], [3, 1, 4, 1], [5, 8, 2], [6, 5, 3], [6, 5, 3], [3, 1, 4, 1], [3, 1, 4, 1]]
    m = eliminate_consecutive_duplicates(l)
    n = [[3, 1, 4, 1], [5, 8, 2], [6, 5, 3], [3, 1, 4, 1]]

    assert_equal(m, n)


def test_set_new_data_basic():

    data_1 = np.random.normal(5, 10, [10, 10, 5])
    affine_1 = np.diag([1, 2, 3, 1])

    data_2 = np.random.normal(5, 10, [3, 2, 4]).astype(np.float32)
    im_data_1 = nib.Nifti1Image(data_1, affine_1)
    im_data_1.set_data_dtype(np.uint8)
    im_data_1.header['descrip'] = 'Spam'

    im_data_2 = set_new_data(im_data_1, data_2)

    assert_array_equal(im_data_2.get_data(), data_2)
    assert_array_equal(im_data_2.affine, affine_1)
    assert_equal(im_data_2.header['descrip'], b'Spam')
    assert_equal(im_data_1.get_data_dtype(), np.uint8)
    assert_equal(im_data_2.get_data_dtype(), np.float32)


def test_set_new_data_new_dtype():

    data_1 = np.random.normal(5, 10, [10, 10, 5])
    affine_1 = np.diag([1, 2, 3, 1])

    data_2 = np.random.normal(5, 10, [3, 2, 4])
    im_data_1 = nib.Nifti1Image(data_1, affine_1)
    im_data_1.set_data_dtype(np.float32)

    im_data_2 = set_new_data(im_data_1, data_2, new_dtype=np.uint16)

    assert_equal(im_data_1.get_data_dtype(), np.float32)
    assert_equal(im_data_2.get_data_dtype(), np.uint16)


def test_set_new_data_nan_no_nan():

    data_1 = np.random.normal(5, 10, [10, 10, 5])
    affine_1 = np.diag([1, 2, 3, 1])

    data_1[2, 2, 2] = np.nan
    data_1[1, 1, 1] = np.nan

    data_2 = np.random.normal(5, 10, [3, 2, 4])

    im_data_1 = nib.Nifti1Image(data_1, affine_1)
    im_data_2 = set_new_data(im_data_1, data_2, new_dtype=np.uint16)

    assert np.nan not in im_data_2.get_data()



import os
import numpy as np
import warnings
import sys

from numpy.testing import assert_array_equal, assert_equal

from bruker2nifti._cores import scan2struct, write_struct

here = os.path.abspath(os.path.dirname(__file__))
root_dir = os.path.dirname(here)


def test_scan2struct_with_banana_data_slope_no_slope():

    pfo_scan_in = os.path.join(root_dir, 'test_data', 'bru_banana', '1')

    struct_no_slope_corrected = scan2struct(pfo_scan_in, correct_slope=False)

    struct_yes_slope_corrected = scan2struct(pfo_scan_in, correct_slope=True)

    parsed_slope_from_no = struct_no_slope_corrected['visu_pars_list'][0]['VisuCoreDataSlope']
    parsed_slope_from_yes = struct_yes_slope_corrected['visu_pars_list'][0]['VisuCoreDataSlope']
    ground_slope = np.array([11.02701887, ] * 5)

    # check general structure:
    assert_equal(struct_no_slope_corrected['nib_scans_list'][0].shape,
                 struct_yes_slope_corrected['nib_scans_list'][0].shape)
    assert_array_equal(parsed_slope_from_no, parsed_slope_from_yes)
    np.testing.assert_array_almost_equal(parsed_slope_from_no, ground_slope)

    assert_equal(struct_no_slope_corrected['acqp'], {})
    assert_equal(struct_no_slope_corrected['reco'], {})
    assert_equal(struct_no_slope_corrected['method'], {})

    # check data of the nibabel images:
    np.testing.assert_array_almost_equal(ground_slope[0] * struct_no_slope_corrected['nib_scans_list'][0].get_data(),
                                         struct_yes_slope_corrected['nib_scans_list'][0].get_data(), decimal=4)


def test_write_struct():

    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter("always")
        # Trigger a warning
        pfo_study_in = os.path.join(root_dir, 'test_data', 'bru_banana')
        scan2struct(pfo_study_in)
        # manage warning in python 2 or python 3:
        if sys.version_info[0] == 3:
            assert len(w) == 1
            assert issubclass(w[-1].category, UserWarning)
        else:
            import exceptions
            assert len(w) == 1
            assert issubclass(w[-1].category, exceptions.UserWarning)

    pfo_scan_in = os.path.join(root_dir, 'test_data', 'bru_banana', '1')
    banana_struct = scan2struct(pfo_scan_in)

    pfo_output = os.path.join(root_dir, 'test_data', 'nifti_banana', 'write_struct_here')
    if not os.path.exists(pfo_output):
        os.system('mkdir {}'.format(pfo_output))

    write_struct(banana_struct, pfo_output, fin_scan='test')

    assert os.path.exists(os.path.join(pfo_output, 'acquisition_method.txt'))
    assert os.path.exists(os.path.join(pfo_output, 'test.nii.gz'))



import argparse
import os

from bruker2nifti.converter import Bruker2Nifti


def main_scan():
    """
    Parser from terminal with
    $ python2 bruker2nifti_scan -h
    $ python2 bruker2nifti_scan -i input_file_path -o output_file_path
    """

    parser = argparse.ArgumentParser()

    # pfo_input_scan
    parser.add_argument('-i', '--input_scan_folder',
                        dest='pfo_input',
                        type=str,
                        required=True,
                        help='Bruker scan folder.')
    # pfo_output
    parser.add_argument('-o', '--output_scan_folder',
                        dest='pfo_output',
                        type=str,
                        required=True,
                        help='Output folder where the study will be saved.')
    # fin_output = None
    parser.add_argument('--fin_output',
                        dest='fin_output',
                        type=str,
                        default=None)
    # nifti_version = 1,
    parser.add_argument('-nifti_version',
                        dest='nifti_version',
                        type=int,
                        default=1,
                        help='Filename of the nifti output.')
    # qform = 2,
    parser.add_argument('-qform_code',
                        dest='qform_code',
                        type=int,
                        default=2)
    # sform = 1,
    parser.add_argument('-sform_code',
                        dest='sform_code',
                        type=int,
                        default=1)

    # do_not_save_human_readable = False,
    parser.add_argument('-do_not_save_human_readable',
                        dest='do_not_save_human_readable',
                        action='store_true')

    # correct_slope = False,
    parser.add_argument('-correct_slope',
                        dest='correct_slope',
                        action='store_true')
    # verbose = 1
    parser.add_argument('-verbose', '-v',
                        dest='verbose',
                        type=int,
                        default=1)

    args = parser.parse_args()

    # instantiate a converter:
    bruconv = Bruker2Nifti(os.path.dirname(args.pfo_input), args.pfo_output)
    # get the attributes
    bruconv.nifti_version = args.nifti_version
    bruconv.qform_code = args.qform_code
    bruconv.sform_code = args.sform_code
    bruconv.save_human_readable = not args.do_not_save_human_readable
    bruconv.correct_slope = args.correct_slope
    bruconv.verbose = args.verbose

    if parser.add_argument > 0:
        print('\nConverter parameters: ')
        print('-------------------------------------------------------- ')
        print('Study Folder         : {}'.format(os.path.dirname(args.pfo_input)))
        print('Scan to convert      : {}'.format(os.path.basename(args.pfo_input)))
        print('List of scans        : {}'.format(bruconv.scans_list))
        print('Output NifTi version : {}'.format(bruconv.nifti_version))
        print('Output NifTi q-form  : {}'.format(bruconv.qform_code))
        print('Output NifTi s-form  : {}'.format(bruconv.sform_code))
        print('Save human readable  : {}'.format(bruconv.save_human_readable))
        print('Correct the slope    : {}'.format(bruconv.correct_slope))
        print('-------------------------------------------------------- ')
    # convert the single:
    bruconv.convert_scan(args.pfo_input, args.pfo_output, nifti_file_name=args.fin_output,
                         create_output_folder_if_not_exists=True)


if __name__ == "__main__":
    main_scan()



import os

from ._utils import bruker_read_files
from ._getters import get_list_scans, get_subject_name
from ._cores import scan2struct, write_struct


class Bruker2Nifti(object):
    """
    Facade to collect users preferences on the conversion and accessing the core methods for the conversion
    (scan_to_struct, write_struct).

    Nomenclature:

    study: a series of acquisition related to the same subject, acquired in the same scanning session and usually
    containing multiple scans. It is provided as a folder structure containing the scans produced with paravision (PV)
    software. Patient/subject information are embedded in the study (opposed hierarchy as in the DICOM files).

    scan or experiment, sub-scans and sub-volumes: individual folder image acquired with various protocols. To a scan
    can belong more than one processed image, or reconstruction. Each processed image can be a single volume or can
    contain more than one sub-volume embedded in the same processed image.

    header: header of the nifti format.

    img_data: data of the nifti format, stored in a 2d or 3d matrix.

    struct: intermediate structure (python dictionary) proposed in this code, aimed at collecting the information from
    the raw Bruker and to progressively creating the nifti images.
    """

    def __init__(self, pfo_study_bruker_input, pfo_study_nifti_output, study_name=None):
        """
        Initialise the Facade class to the converter.
        :param pfo_study_bruker_input: path to folder (pfo) to the Bruker input data folder.
        :param pfo_study_nifti_output: path to folder (pfo) where the converted study will be stored.
        :param study_name: optional name of the study. If None, the name parsed from the Bruker study will be used.
        """
        self.pfo_study_bruker_input = pfo_study_bruker_input
        self.pfo_study_nifti_output = pfo_study_nifti_output
        self.study_name             = study_name
        # converter settings for the nifti values
        self.nifti_version       = 1
        self.qform_code          = 1
        self.sform_code          = 2
        self.save_human_readable = True
        self.save_b0_if_dwi      = True  # if DWI, it saves the first layer as a single nfti image.
        self.correct_slope       = True
        self.correct_offset      = True
        # advanced sample positioning
        self.sample_upside_down       = False
        self.frame_body_as_frame_head = False
        # chose to convert extra files:
        self.get_acqp   = False
        self.get_method = False
        self.get_reco   = False
        # advanced selections:
        self.scans_list              = None  # you can select a subset of scans in the study to be converted.
        self.list_new_name_each_scan = None  # you can select specific names for the subset self.scans_list.
        self.verbose                 = 1
        # automatic filling of advanced selections class attributes
        self.explore_study()

    def explore_study(self):
        """
        Automatic filling of the advanced selections class attributes.
        It also checks if the given attributes are meaningful.
        :return:
        """

        if not os.path.isdir(self.pfo_study_bruker_input):
            raise IOError('Input folder does not exist.')
        if not os.path.isdir(self.pfo_study_nifti_output):
            raise IOError('Output folder does not exist.')
        if self.scans_list is None:
            self.scans_list = get_list_scans(self.pfo_study_bruker_input, print_structure=False)
            assert isinstance(self.scans_list, list)
            msg = 'No scans found, are you sure the input folder contains a Bruker study?'
            if not len(self.scans_list) > 0:
                raise IOError(msg)
        if self.study_name is None or self.study_name is '':
            _study_name = get_subject_name(self.pfo_study_bruker_input).replace(' ', '_')
            self.study_name = ''.join(e for e in _study_name if e.isalnum())
        if self.list_new_name_each_scan is None:
            list_new_name_each_scan = [self.study_name + '_' + ls for ls in self.scans_list]
            self.list_new_name_each_scan = list_new_name_each_scan
            assert isinstance(self.list_new_name_each_scan, list)
        # if self.list_new_nifti_file_names is None:
        #     self.list_new_nifti_file_names = self.list_new_name_each_scan
        else:
            if not len(self.scans_list) == len(self.list_new_name_each_scan):
                msg = 'list_name_each_scan {0} does not have the same amount of scans in the ' \
                      'study: {1}'.format(self.list_new_name_each_scan, self.scans_list)
                raise IOError(msg)

    def show_study_structure(self):
        """
        Print to console the structure of the selected study.
        :return: [None] only print to console information.
        """
        if not os.path.isdir(self.pfo_study_bruker_input):
            raise IOError('Input folder does not exist.')

        print('Study folder structure: ')
        scans_list = get_list_scans(self.pfo_study_bruker_input)
        print('\n')
        print('List of scans: {}'.format(scans_list))
        pfi_first_scan = os.path.join(self.pfo_study_bruker_input, scans_list[0])
        acqp = bruker_read_files('acqp', pfi_first_scan)
        print('Version: {}'.format(acqp['ACQ_sw_version'][0]))

    def convert_scan(self, pfo_input_scan, pfo_output_converted, nifti_file_name=None,
                     create_output_folder_if_not_exists=True):
        """
        :param pfo_input_scan: path to folder (pfo) containing a scan from Bruker, see documentation for the difference
         between Bruker 'scan' and Bruker 'study'.
        :param pfo_output_converted: path to the folder where the converted scan will be stored.
        :param create_output_folder_if_not_exists: [True] if the output folder does not exist will be created.
        :param nifti_file_name: [None] filename of the nifti image that will be saved into the pfo_output folder.
         If None, the filename will be obtained from the parameter file of the study.
        :return: [None] save the data parsed from the raw Bruker scan into a folder, including the nifti image.
        """

        if not os.path.isdir(pfo_input_scan):
            raise IOError('Input folder does not exist.')

        if create_output_folder_if_not_exists:
            if 'win' in os.name.lower():
                os.system('mkdir {}'.format(pfo_output_converted))
            else:
                os.system('mkdir -p {}'.format(pfo_output_converted))

        print('FRAME BODY {}'.format(self.frame_body_as_frame_head))
        print('UPSIDE DOWN {}'.format(self.sample_upside_down))

        struct_scan = scan2struct(pfo_input_scan,
                                  correct_slope=self.correct_slope,
                                  correct_offset=self.correct_offset,
                                  sample_upside_down=self.sample_upside_down,
                                  nifti_version=self.nifti_version,
                                  qform_code=self.qform_code,
                                  sform_code=self.sform_code,
                                  get_acqp=self.get_acqp,
                                  get_method=self.get_method,
                                  get_reco=self.get_reco,
                                  frame_body_as_frame_head=self.frame_body_as_frame_head
                                  )

        if struct_scan is not None:
            write_struct(struct_scan,
                         pfo_output_converted,
                         fin_scan=nifti_file_name,
                         save_human_readable=self.save_human_readable,
                         save_b0_if_dwi=self.save_b0_if_dwi,
                         verbose=self.verbose,
                         )

    def convert(self):
        """
        To call the converter, once all the settings of the converter are selected and modified by the user.
        :return: Convert the Bruker study, whose path is stored in the class variable self.pfo_study_bruker_input
        in the specified folder stored in self.pfo_study_nifti_output, and according to the other class attributes.

        Example:

        >> bru = Bruker2Nifti('/path/to/my/study', '/path/output', study_name='study1')

        >> bru.show_study_structure

        >> bru.verbose = 2
        >> bru.correct_slope = True
        >> bru.get_acqp = False
        >> bru.get_method = True  # I want to see the method parameter file converted as well.
        >> bru.get_reco = False

        >> # Convert the study:
        >> bru.convert()

        """
        pfo_nifti_study = os.path.join(self.pfo_study_nifti_output, self.study_name)
        os.system('mkdir -p {0}'.format(pfo_nifti_study))

        print('\nStudy conversion \n{}\nstarted:\n'.format(self.pfo_study_bruker_input))

        for bruker_scan_name, scan_name in zip(self.scans_list, self.list_new_name_each_scan):
            pfo_scan_bruker = os.path.join(self.pfo_study_bruker_input, bruker_scan_name)
            pfo_scan_nifti = os.path.join(pfo_nifti_study, scan_name)

            print('\nConverting experiment {}:\n'.format(bruker_scan_name))

            self.convert_scan(pfo_scan_bruker,
                               pfo_scan_nifti,
                               create_output_folder_if_not_exists=True,
                               nifti_file_name=scan_name)

        print('\nStudy converted and saved in \n{}'.format(self.pfo_study_nifti_output))



from subprocess import check_output
import os


# # Describe the version relative to last tag
# command_git = ['git', 'describe', '--match', 'v[0-9]*']
# version_buf = check_output(command_git).rstrip()
#
# print(version_buf)
#
# # Exclude the 'v' for PEP440 conformity, see
# # https://www.python.org/dev/peps/pep-0440/#public-version-identifiers
# version_buf = version_buf[1:]
#
# print(version_buf)
import numpy as np
import os

from bruker2nifti.converter import Bruker2Nifti
from bruker2nifti._getters import get_stack_direction_from_VisuCorePosition


if __name__ == '__main__':
    root_dir = '/Users/sebastiano/a_data'
    pfo_study_in = os.path.join(root_dir, 'bruker', '3103')
    pfo_study_out = os.path.join(root_dir, 'bruker', 'converted')

    print(os.path.isdir(pfo_study_out))

    # # instantiate a converter
    # bru = Bruker2Nifti(pfo_study_in, pfo_study_out, study_name='my_study')
    # # select the options (attributes) you may want to change - the one shown below are the default one:
    # bru.verbose = 2
    # bru.correct_slope = True
    # bru.get_acqp = False
    # bru.get_method = False
    # bru.get_reco = False
    # bru.nifti_version = 1
    # bru.qform_code = 1
    # bru.sform_code = 2
    # bru.save_human_readable = True
    # bru.save_b0_if_dwi = True
    # # Check that the list of scans and the scans names automatically selected makes some sense:
    # print(bru.scans_list)
    # print(bru.list_new_name_each_scan)
    # # call the function convert, to convert the study:
    # bru.convert()
    '''
    visu_pars = np.load(os.path.join(pfo_study_out, 'my_study', 'my_study_8', 'my_study_8_visu_pars.npy'))

    vp = visu_pars.item()

    vp['VisuCorePosition'].shape



    get_stack_direction_from_VisuCorePosition(vp['VisuCorePosition'])
    '''
    # visu_core_position_ = np.array([[-20., -20.,  -4.],
    #                                 [-20., -20.,  -2.],
    #                                 [-20., -20.,   0.],
    #                                 [-20., -20.,   2.],
    #                                 [-20., -20.,   4.],
    #                                 [4., -20.,  20.],
    #                                 [2., -20.,  20.],
    #                                 [0., -20.,  20.],
    #                                 [-2., -20.,  20.],
    #                                 [-4., -20.,  20.],
    #                                 [-20.,  4., 20.],
    #                                 [-20.,  2., 20.],
    #                                 [-20., 0., 20.],
    #                                 [-20., - 2.,  20.],
    #                                 [-20., - 4., 20.]])
    # a = get_stack_direction_from_VisuCorePosition(visu_core_position_, 3)
    # print a

    vcp = np.array([[-4., -20., 20.]])
    get_stack_direction_from_VisuCorePosition(vcp, 1)



import sys
import argparse
from bruker2nifti.converter import Bruker2Nifti


def main():
    """
    Parser from terminal with:
    $ python2 bruker2nifti -h
    $ python2 bruker2nifti -i input_file_path -o output_file_path
    """

    parser = argparse.ArgumentParser()

    # custom helper
    parser.add_argument('-what',
                        dest='what',
                        action='store_true',
                        required=False,
                        help='Get more information about the software')

    # pfo_study_bruker_input
    parser.add_argument('-i', '--input_study_folder',
                        dest='pfo_input',
                        type=str,
                        required=False,
                        help='Bruker study folder.')

    # pfo_study_nifti_output
    parser.add_argument('-o', '--output_study_folder',
                        dest='pfo_output',
                        type=str,
                        required=False,
                        help='Output folder where the study will be saved.')

    # study_name = None,
    parser.add_argument('-study_name',
                        dest='study_name',
                        default=None)

    # scans_list = None
    parser.add_argument('-scans_list',
                        dest='scans_list',
                        default=None)

    # list_new_name_each_scan = None,
    parser.add_argument('-list_new_name_each_scan',
                        dest='list_new_name_each_scan',
                        default=None)

    # nifti_version = 1,
    parser.add_argument('-nifti_version',
                        dest='nifti_version',
                        type=int,
                        default=1)

    # qform = 2,
    parser.add_argument('-qform_code',
                        dest='qform_code',
                        type=int,
                        default=1)

    # sform= 1,
    parser.add_argument('-sform_code',
                        dest='sform_code',
                        type=int,
                        default=2)

    # do_not_save_human_readable = False,
    parser.add_argument('-do_not_save_human_readable',
                        dest='do_not_save_human_readable',
                        action='store_true')

    # correct_slope = True,
    parser.add_argument('-correct_slope',
                        dest='correct_slope',
                        action='store_true')

    # correct_offset = True,
    parser.add_argument('-correct_offset',
                        dest='correct_offset',
                        action='store_true')

    # correct_offset = True,
    parser.add_argument('-sample_upside_down',
                        dest='sample_upside_down',
                        action='store_false')

    # correct_offset = True,
    parser.add_argument('-frame_body_as_frame_head',
                        dest='frame_body_as_frame_head',
                        action='store_false')

    # verbose = 1
    parser.add_argument('-verbose', '-v',
                        dest='verbose',
                        type=int,
                        default=1)

    # ------ Parsing user's input ------ #

    args = parser.parse_args()

    # Check input:
    if args.what:
        msg = 'Code repository : {} \n' \
              'Documentation   : {}'.format('https://github.com/SebastianoF/bruker2nifti',
                                            'https://github.com/SebastianoF/bruker2nifti/wiki')
        sys.exit(msg)

    if not args.pfo_input or not args.pfo_output:
        sys.exit('Input bruker study [-i] and output folder [-o] required')

    # Instantiate a converter:
    bruconv = Bruker2Nifti(args.pfo_input,
                           args.pfo_output,
                           study_name=args.study_name)

    if args.scans_list is not None:
        bruconv.scans_list = args.scans_list
    if args.list_new_name_each_scan is not None:
        bruconv.list_new_name_each_scan = args.list_new_name_each_scan

    # Basics
    bruconv.nifti_version       = args.nifti_version
    bruconv.qform_code          = args.qform_code
    bruconv.sform_code          = args.sform_code
    bruconv.save_human_readable = not args.do_not_save_human_readable
    bruconv.correct_slope       = args.correct_slope
    bruconv.correct_offset      = args.correct_offset
    bruconv.verbose             = args.verbose
    # Sample position
    bruconv.sample_upside_down       = args.sample_upside_down
    bruconv.frame_body_as_frame_head = args.frame_body_as_frame_head

    print('\nConverter input parameters: ')
    print('-------------------------------------------------------- ')
    print('Study name           : {}'.format(bruconv.study_name))
    print('List of scans        : {}'.format(bruconv.scans_list))
    print('List of scans names  : {}'.format(bruconv.list_new_name_each_scan))
    print('Output NifTi version : {}'.format(bruconv.nifti_version))
    print('Output NifTi q-form  : {}'.format(bruconv.qform_code))
    print('Output NifTi s-form  : {}'.format(bruconv.sform_code))
    print('Save human readable  : {}'.format(bruconv.save_human_readable))
    print('Correct the slope    : {}'.format(bruconv.correct_slope))
    print('Correct the offset   : {}'.format(bruconv.correct_offset))
    print('-------------------------------------------------------- ')
    print('Sample upside down         : {}'.format(bruconv.sample_upside_down))
    print('Frame body as frame head   : {}'.format(bruconv.frame_body_as_frame_head))
    print('-------------------------------------------------------- ')
    bruconv.convert()


if __name__ == "__main__":
    main()



import os

try:
    # for Python2
    import Tkinter as tk
    import tkFileDialog
except ImportError:
    # for Python3
    import tkinter as tk
    from tkinter import filedialog as tkFileDialog

import webbrowser

from bruker2nifti.converter import Bruker2Nifti
from bruker2nifti.__init__ import __version__ as version


class BrukerToNiftiGUI(tk.Tk, object):
    """
    Graphical user interface class to access Bruker to Nifti converter.
    """
    def __init__(self, in_pfo_input=None, in_pfo_output=None, in_study_name=None):

        super(BrukerToNiftiGUI, self).__init__()

        # Window settings:

        self.title('From bruker to nifti - interface - version {}'.format(version))
        self.geometry('715x225')

        # Widgets:

        self.header = tk.Label(self, text=' Bruker 2 nifti Converter ', font="Verdana 14")

        self.label_pfo_input = tk.Label(self, text='Folder path input')
        self.entry_pfo_input = tk.Entry(self, bd=2, width=50)

        if in_pfo_input is not None:
            self.entry_pfo_input.insert(0, in_pfo_input)

        self.label_pfo_output = tk.Label(self, text='Folder path output')
        self.entry_pfo_output = tk.Entry(self, bd=2, width=50)

        if in_pfo_output is not None:
            self.entry_pfo_output.insert(0, in_pfo_output)

        self.label_study_name = tk.Label(self, text='Study name (optional)')
        self.entry_study_name = tk.Entry(self, bd=2, width=50)

        if in_study_name is not None:
            self.entry_study_name.insert(0, in_study_name)

        self.CheckVar_cs = tk.IntVar(value=1)  # correct slope default True
        self.CheckVar_co = tk.IntVar(value=1)  # correct offset default True
        self.CheckVar_ac = tk.IntVar(value=0)  # get acquisition parameters default False
        self.CheckVar_me = tk.IntVar(value=0)  # get method default False
        self.CheckVar_rc = tk.IntVar(value=0)  # get reco default False
        self.CheckVar_ud = tk.IntVar(value=0)  # sample upside down, default False

        self.radio_button_correct_slope  = tk.Checkbutton(self, text='correct slope', variable=self.CheckVar_cs, compound='right')
        self.radio_button_correct_offset = tk.Checkbutton(self, text='correct offset', variable=self.CheckVar_co, compound='right')
        self.radio_button_get_acqp       = tk.Checkbutton(self, text='get acqp', variable=self.CheckVar_ac, compound='right')
        self.radio_button_get_method     = tk.Checkbutton(self, text='get method', variable=self.CheckVar_me, compound='right')
        self.radio_button_get_reco       = tk.Checkbutton(self, text='get reco', variable=self.CheckVar_rc, compound='right')

        self.radio_button_sample_ups_down = tk.Checkbutton(self, text='sample upside down', variable=self.CheckVar_ud, compound='right')

        self.button_browse_input = tk.Button(self, text='Browse', command=self.button_browse_callback_pfo_input)
        self.button_browse_output = tk.Button(self, text='Browse', command=self.button_browse_callback_pfo_output)
        self.button_help = tk.Button(self, text='?', command=self.open_help, fg='blue', font="times 14 bold")
        self.button_convert = tk.Button(self, text='Convert', command=self.convert, highlightbackground="#1291E9")

        self.label_option_nifti = tk.Label(self, text='Output NifTi version:', compound='right')
        self.option_menu_nifti_value = tk.StringVar(self)
        self.option_menu_nifti_value.set('1')
        self.option_menu_nifti = tk.OptionMenu(self, self.option_menu_nifti_value, '1', '2')

        self.label_option_frame = tk.Label(self, text='Frame orientation:', compound='right')
        self.option_menu_frame_value = tk.StringVar(self)
        self.option_menu_frame_value.set('Rodent')
        self.option_menu_frame = tk.OptionMenu(self, self.option_menu_frame_value, 'Rodent', 'Primate')

        # geometry

        self.header.grid(row=0, column=1, columnspan=3)

        self.label_pfo_input.grid(row=1, column=0)
        self.entry_pfo_input.grid(row=1, column=1, columnspan=3)
        self.button_browse_input.grid(row=1, column=4)

        self.label_pfo_output.grid(row=2, column=0)
        self.entry_pfo_output.grid(row=2, column=1, columnspan=3)
        self.button_browse_output.grid(row=2, column=4)

        self.label_study_name.grid(row=3, column=0)
        self.entry_study_name.grid(row=3, column=1, columnspan=3)

        self.radio_button_correct_slope.grid(row=4, column=0)
        self.radio_button_correct_offset.grid(row=4, column=1)
        self.radio_button_sample_ups_down.grid(row=4, column=2)

        self.radio_button_get_acqp.grid(row=5, column=0)
        self.radio_button_get_method.grid(row=5, column=1)
        self.radio_button_get_reco.grid(row=5, column=2)

        self.button_convert.grid(row=7, column=4)
        self.button_help.grid(row=0, column=4)

        self.label_option_nifti.grid(row=6, column=0)
        self.option_menu_nifti.grid(row=6, column=1)

        self.label_option_frame.grid(row=7, column=0)
        self.option_menu_frame.grid(row=7, column=1)

    # main commands

    def button_browse_callback_pfo_input(self):
        filename = tkFileDialog.askdirectory()
        filename = os.path.abspath(filename)
        self.entry_pfo_input.delete(0, tk.END)
        self.entry_pfo_input.insert(0, filename)

    def button_browse_callback_pfo_output(self):
        filename = tkFileDialog.askdirectory()
        filename = os.path.abspath(filename)
        self.entry_pfo_output.delete(0, tk.END)
        self.entry_pfo_output.insert(0, filename)

    @staticmethod
    def open_help():
        webbrowser.open('https://github.com/SebastianoF/bruker2nifti/wiki/Graphical-User-Interface-Examples')

    def convert(self):

        bru = Bruker2Nifti(self.entry_pfo_input.get(),
                           self.entry_pfo_output.get(),
                           study_name=self.entry_study_name.get())

        bru.correct_slope  = self.CheckVar_cs.get()
        bru.correct_offset = self.CheckVar_co.get()
        bru.get_acqp       = self.CheckVar_ac.get()
        bru.get_method     = self.CheckVar_me.get()
        bru.get_reco       = self.CheckVar_rc.get()
        bru.nifti_version  = int(self.option_menu_nifti_value.get())

        bru.sample_upside_down = self.CheckVar_ud.get()
        rodent_primate         = self.option_menu_frame_value.get()

        bru.frame_body_as_frame_head = False
        if rodent_primate == 'Primate':
            bru.frame_body_as_frame_head = True

        print(' --- GUI bruker2nifti --- ')
        print('Input path     : {}'.format(bru.pfo_study_bruker_input))
        print('Output path    : {}'.format(bru.pfo_study_nifti_output))
        print('Study name     : {}'.format(bru.study_name))
        print(' ------------------ ')
        print('Correct slope  : {}'.format(bru.correct_slope))
        print('Correct offset : {}'.format(bru.correct_offset))
        print('Get acpq       : {}'.format(bru.get_acqp))
        print('get method     : {}'.format(bru.get_method))
        print('get reco       : {}'.format(bru.get_reco))
        print('get NifTi-     : {}'.format(bru.nifti_version))
        print(' ------------------ ')
        print('Sample upside down       : {}'.format(bru.sample_upside_down))
        print('Frame body as frame head : {}'.format(bru.frame_body_as_frame_head))

        bru.convert()

        del bru

        print('\nbruker2Nifti verision {} - https://github.com/SebastianoF/bruker2nifti'.format(version))


def open_gui(in_pfo_input=None, in_pfo_output=None, in_study_name=None):
    """
    To open the Graphical user interface accessing Bruker2Nifti converter.
    :param in_pfo_input: [None] input path to folder (pfo) where the bruker folder structure is located.
    It that will initialise the corresponding field entry on the GUI.
    :param in_pfo_output: [None] input path to folder (pfo) where the corresponding folder structure with nifti images
    and selected parameter files will be located.
    :param in_study_name: [None] to initialise the optional study name field.
    :return: Will open the gui.
    """
    root = BrukerToNiftiGUI(in_pfo_input=in_pfo_input, in_pfo_output=in_pfo_output, in_study_name=in_study_name)
    root.mainloop()


if __name__ == '__main__':
    open_gui()



import os


from bruker2nifti.converter import Bruker2Nifti

if __name__ == '__main__':
    pfo_in = '/Users/sebastiano/Downloads/z_test_scan'
    pfo_out = '/Users/sebastiano/Downloads/out'
    print(os.path.isdir(pfo_out))
    # instantiate a converter
    bru = Bruker2Nifti(pfo_in, pfo_out, study_name=None)
    # select the options (attributes) you may want to change - the one shown below are the default one:
    bru.verbose = 2
    bru.correct_slope = True
    bru.correct_offset = False
    bru.get_acqp = True
    bru.get_method = True
    bru.get_reco = True
    bru.nifti_version = 1
    bru.qform_code = 1
    bru.sform_code = 2
    bru.save_human_readable = True
    bru.save_b0_if_dwi = True

    # Check that the list of scans and the scans names automatically selected makes some sense:
    print(bru.scans_list)
    print(bru.list_new_name_each_scan)
    print(bru.study_name)
    # call the function convert, to convert the study:
    bru.convert()




import os
import nibabel as nib
import numpy as np

from ._utils import bruker_read_files, eliminate_consecutive_duplicates, data_corrector, \
    compute_affine_from_visu_pars, compute_resolution_from_visu_pars


def get_list_scans(start_path, print_structure=True):
    """
    Given a path containing scans (pfo_study) or sub-scans (join(pfo_study, 'pdata')),
    finds the list of the names of the scans.
    :param start_path: path to the folder containing scans or sub-scans.
    :param print_structure: [True] optional if you want to visualise the structure data at console
    :return: list of scans/sub_scans names.
    """

    scans_list = []

    for dirpath, dirnames, filenames in os.walk(start_path):

        if dirpath == start_path:
            scans_list = [d for d in dirnames if d.isdigit()]

        level = dirpath.replace(start_path, '').count(os.sep)
        indent = (' ' * 4) * level
        if print_structure:
            print('{}{}/'.format(indent, os.path.basename(dirpath)))

        sub_indent = (' ' * 4) * (level + 1)

        if print_structure:
            for f in filenames:
                print('{}{}'.format(sub_indent, f))

    scans_list.sort(key=float)
    return scans_list


def get_subject_name(pfo_study):
    """
    :param pfo_study: path to study folder.
    :return: name of the subject in the study. See get_subject_id.
    """
    # (1) 'subject' at the study level is present
    if os.path.exists(os.path.join(pfo_study, 'subject')):
        subject = bruker_read_files('subject', pfo_study)
        return subject['SUBJECT_id']
    # (2) 'subject' at the study level is not present, we use 'VisuSubjectId' from visu_pars of the first scan.
    # Longer solution as at the end 'visu_pars' will be unavoidably scanned twice.
    else:
        list_scans = get_list_scans(pfo_study)
        visu_pars = bruker_read_files('visu_pars', pfo_study, sub_scan_num=list_scans[0])
        return visu_pars['VisuSubjectId']


def get_stack_direction_from_VisuCorePosition(visu_core_position, num_sub_volumes=1):
    """
    To be used when the acquisition is 2D: it returns the stack orientation and direction encoded in a string.

    :param visu_core_position: visu_pars parameter file parsed into a dictionary.
    :param num_sub_volumes: [1] for when there is more than one volume embedded in the same scan
    (see num_vols variable of the method nifti_getter)
    :return: stack direction and axis of the 2D image. If more than one sub-volume is included, returns
    the sequence of string-encoded orientations and directions.
    ----------
    Example:

    visu_core_position -> corresponding output:
    [[  -4. -20.  20.] [ -2. -20.  20.] [ 0. -20.  20.] [ 2. -20.  20.] [ 4. -20.  20.]]    -> 'x+'
    [[  4. -20.  20.]  [ 2. -20.  20.]   [ 0. -20.  20.] [ -2. -20.  20.]  [-4. -20.  20.]] -> 'x-'
    [[-20.  -4.  20.] [-20.  -2.  20.] [-20.   0.  20.] [-20.  2.  20.] [-20.  4.  20.]]    -> 'y+'
    [[-20.   4.  20.] [-20.   2.  20.] [-20.   0.  20.] [-20.  -2.  20.] [-20.  -4.  20.]]  -> 'y-'
    [[-20. -20.  -4.] [-20. -20.  -2.] [-20. -20.   0.] [-20. -20.   2.] [-20. -20.   4.]]  -> 'z+'
    [[-20. -20.  4.] [-20. -20.  2.] [-20. -20.   0.] [-20. -20.   -2.] [-20. -20.   -4.]]  -> 'z-'

    visu_core_position = [[-20. -20.  -4.]
                          [-20. -20.  -2.]
                          [-20. -20.   0.]
                          [-20. -20.   2.]
                          [-20. -20.   4.]
                          [  4. -20.  20.]
                          [  2. -20.  20.]
                          [  0. -20.  20.]
                          [ -2. -20.  20.]
                          [ -4. -20.  20.]]
    num_sub_volumes = 2 -> 'z-x-'

    """
    sh = visu_core_position.shape
    if not len(sh) == 2:
        msg = 'The input VisuCorePosition must be from a 2D acquisition. Input shape {}'.format(sh)
        raise IOError(msg)
    if not sh[0] > 1:
        msg = 'The input VisuCorePosition must be from a 2D acquisition. Needs to have more than one row. ' \
              'Input shape {}'.format(sh)
        raise IOError(msg)

    rows, cols = sh
    if not cols == 3:
        raise IOError('VisuCorePosition should have three columns.')

    slices_per_vol, reminder = int(rows // num_sub_volumes), rows % num_sub_volumes
    if not reminder == 0:
        raise IOError('Number of subvolumes not compatible with VisuCorePosition.')

    res = ''
    for v in range(num_sub_volumes):
        first_slice_each_vol = slices_per_vol * v
        s_diff = visu_core_position[first_slice_each_vol + 1, :] - visu_core_position[first_slice_each_vol, :]
        pos = [i for i in range(3) if np.abs(s_diff[i]) > 0][0]
        val = s_diff[pos]
        if pos == 0:
            res += 'x'
        elif pos == 1:
            res += 'y'
        elif pos == 2:
            res += 'z'
        if val > 0:
            res += '+'
        else:
            res += '-'
    return res


def nifti_getter(img_data_vol,
                 visu_pars,
                 correct_slope,
                 correct_offset,
                 sample_upside_down,
                 nifti_version,
                 qform_code,
                 sform_code,
                 frame_body_as_frame_head=False,
                 keep_same_det=True,
                 consider_subject_position=False
                 ):
    """
    Passage method to get a nifti image from the volume and the element contained into visu_pars.
    :param img_data_vol: volume of the image.
    :param visu_pars: corresponding dictionary to the 'visu_pars' data file.
    :param correct_slope: [True/False] if you want to correct the slope.
    :param correct_offset: [True/False] if you want to correct the offset.
    :param sample_upside_down: [True/False] if you want to have the sample rotated 180 around the Ant-Post axis.
    :param nifti_version: [1/2] according to the required nifti output
    :param qform_code: required nifti ouptut qform code.
    :param sform_code: required nifti ouput sform code
    :param frame_body_as_frame_head: [True/False] if the frame is the same for head and body [monkey] or not [mouse].
    :param keep_same_det: flag to constrain the determinant to be as the one provided into the orientation parameter.
    :param consider_subject_position: [False] if taking into account the 'Head_prone' 'Head_supine' input.
    :return:
    """
    # Check units of measurements:
    if not ['mm', ] * len(visu_pars['VisuCoreSize']) == visu_pars['VisuCoreUnits']:
        # if the UoM is not mm, change here. Add other measurements and refer to xyzt_units from nibabel convention.
        print('Warning, measurement not in mm. This version of the converter deals with data in mm only.')

    # get pre-shape and re-shape volume: (pre-shape is the shape compatible with the slope).
    vol_pre_shape = [int(i) for i in visu_pars['VisuCoreSize']]
    if int(visu_pars['VisuCoreFrameCount']) > 1:
        vol_pre_shape += [int(visu_pars['VisuCoreFrameCount'])]

    if np.prod(vol_pre_shape) == img_data_vol.shape[0]:
        vol_data = img_data_vol.reshape(vol_pre_shape, order='F')
    else:
        echo = img_data_vol.shape[0] / np.prod(vol_pre_shape)
        vol_pre_shape += [echo]
        vol_pre_shape = [int(k) for k in vol_pre_shape]
        vol_data = img_data_vol.reshape(vol_pre_shape, order='F')

    # correct slope if required
    if correct_slope:
        vol_data = data_corrector(vol_data, visu_pars['VisuCoreDataSlope'], kind='slope')
    # correct offset (AFTER slope) if required
    if correct_offset:
        vol_data = data_corrector(vol_data, visu_pars['VisuCoreDataOffs'], kind='offset')

    # get number sub-volumes
    num_sub_volumes = len(eliminate_consecutive_duplicates(list(visu_pars['VisuCoreOrientation'])))

    if num_sub_volumes > 1:

        output_nifti = []

        assert vol_pre_shape[2] % num_sub_volumes == 0
        vol_shape = vol_pre_shape[0], vol_pre_shape[1], int(vol_pre_shape[2] / num_sub_volumes)

        # get resolution - same for all sub-volumes.
        resolution = compute_resolution_from_visu_pars(visu_pars['VisuCoreExtent'],
                                                       visu_pars['VisuCoreSize'],
                                                       visu_pars['VisuCoreFrameThickness'])

        for id_sub_vol in range(num_sub_volumes):

            # compute affine
            affine_transf = compute_affine_from_visu_pars(
                                   list(visu_pars['VisuCoreOrientation'])[id_sub_vol * vol_shape[2]],
                                   list(visu_pars['VisuCorePosition'])[id_sub_vol * vol_shape[2]],
                                   visu_pars['VisuSubjectPosition'],
                                   resolution,
                                   frame_body_as_frame_head=frame_body_as_frame_head,
                                   keep_same_det=keep_same_det,
                                   consider_subject_position=consider_subject_position)

            if sample_upside_down:
                affine_transf = affine_transf.dot(np.diag([-1, 1, -1, 1]))

            # get sub volume in the correct shape
            img_data_sub_vol = vol_data[..., id_sub_vol * vol_shape[2] : (id_sub_vol + 1) * vol_shape[2]]

            if nifti_version == 1:
                nib_im_sub_vol = nib.Nifti1Image(img_data_sub_vol, affine=affine_transf)
            elif nifti_version == 2:
                nib_im_sub_vol = nib.Nifti2Image(img_data_sub_vol, affine=affine_transf)
            else:
                raise IOError('Nifti versions allowed are 1 or 2.')

            hdr_sub_vol = nib_im_sub_vol.header
            hdr_sub_vol.set_qform(affine_transf, code=qform_code)
            hdr_sub_vol.set_sform(affine_transf, code=sform_code)
            hdr_sub_vol['xyzt_units'] = 10  # default mm, seconds
            nib_im_sub_vol.update_header()

            output_nifti.append(nib_im_sub_vol)

    else:

        # get shape
        sh = vol_pre_shape

        # check for frame groups:  -- Very convoluted scaffolding. Waiting to have more infos to refactor this part.
        # ideally an external function read VisuFGOrderDesc should provide the sh and the choice between # A and # B
        # while testing for exception.

        if 'VisuFGOrderDescDim' in visu_pars.keys():  # see manuals D-2-73
            if visu_pars['VisuFGOrderDescDim'] > 0:
                if isinstance(visu_pars['VisuFGOrderDesc'], list):
                    if len(visu_pars['VisuFGOrderDesc']) > 1:
                        descr = visu_pars['VisuFGOrderDesc'][:]
                        # sort descr so that FG_SLICE is the first one, all the others came as they are after swapping.
                        fg_slice_pos = -1
                        fg_echo = -1
                        fg_movie = -1
                        for d in range(len(descr)):
                            if '<FG_SLICE>' in descr[d]:
                                fg_slice_pos = d
                            if '<FG_ECHO>' in descr[d]:
                                fg_echo = d
                            if '<FG_MOVIE>' in descr[d]:
                                fg_movie = d
                        if fg_slice_pos == -1:
                            raise IOError('FG_SLICE not found in the order descriptor, can not tell the ordering.')

                        descr[fg_slice_pos], descr[0] = descr[0], descr[fg_slice_pos]

                        dims = []
                        for dd in descr:
                            dims.append(int(dd.replace('(', '').replace(')', '').split(',')[0]))

                        if np.prod(dims) == sh[-1]:
                            sh = vol_pre_shape[:-1] + dims
                            # A
                            if fg_echo > -1:
                                # MSME
                                stack_data = np.zeros(sh, dtype=vol_data.dtype)
                                for t in range(sh[3]):
                                    for z in range(sh[2]):
                                        stack_data[:, :, z, t] = vol_data[:, :, z * sh[3] + t]

                                vol_data = np.copy(stack_data)
                            # B
                            elif fg_movie > -1:
                                # DTI
                                vol_data = vol_data.reshape(sh, order='F')
                            else:
                                # Else ?
                                vol_data = vol_data.reshape(sh, order='F')

        # get resolution
        resolution = compute_resolution_from_visu_pars(visu_pars['VisuCoreExtent'],
                                                       visu_pars['VisuCoreSize'],
                                                       visu_pars['VisuCoreFrameThickness'])

        # compute affine
        affine_transf = compute_affine_from_visu_pars(list(visu_pars['VisuCoreOrientation'])[0],
                                                      list(visu_pars['VisuCorePosition'])[0],
                                                      visu_pars['VisuSubjectPosition'],
                                                      resolution,
                                                      frame_body_as_frame_head=frame_body_as_frame_head,
                                                      keep_same_det=keep_same_det,
                                                      consider_subject_position=consider_subject_position)

        if sample_upside_down:
            affine_transf = affine_transf.dot(np.diag([-1, 1, -1, 1]))

        if nifti_version == 1:
            output_nifti = nib.Nifti1Image(vol_data, affine=affine_transf)
        elif nifti_version == 2:
            output_nifti = nib.Nifti2Image(vol_data, affine=affine_transf)
        else:
            raise IOError('Nifti versions allowed are 1 or 2.')

        hdr_sub_vol = output_nifti.header
        hdr_sub_vol.set_qform(affine_transf, code=qform_code)
        hdr_sub_vol.set_sform(affine_transf, code=sform_code)
        hdr_sub_vol['xyzt_units'] = 10
        output_nifti.update_header()

    return output_nifti



import os

import warnings
import subprocess
import platform

from bruker2nifti.converter import Bruker2Nifti


here = os.path.abspath(os.path.dirname(__file__))
root_dir = os.path.dirname(here)


def test_convert_the_banana(open_converted=False):

    pfo_study_in = os.path.join(root_dir, 'test_data', 'bru_banana')
    pfo_study_out = os.path.join(root_dir, 'test_data', 'nifti_banana')

    # delete study if already exists:
    target_folder = os.path.join(pfo_study_out, 'banana')
    if os.path.exists(target_folder):
        os.system('rm -r {}'.format(os.path.join(target_folder)))

    # instantiate the converter:
    bru = Bruker2Nifti(pfo_study_in, pfo_study_out, study_name='banana')

    bru.verbose = 2
    bru.correct_slope = True
    bru.get_acqp = False
    bru.get_method = False
    bru.get_reco = False

    bru.convert()

    if open_converted:

        if platform.system() == "Windows":
            os.startfile(pfo_study_out.encode('string-escape'))
        elif platform.system() == "Darwin":
            subprocess.Popen(["open", pfo_study_out])
        else:
            subprocess.Popen(["xdg-open", pfo_study_out])

    for ex in ['1', '2', '3']:
        experiment_folder = os.path.join(pfo_study_out, 'banana', 'banana_{}'.format(ex))
        assert os.path.exists(experiment_folder)
        assert os.path.exists(os.path.join(experiment_folder, 'banana_{}.nii.gz'.format(ex)))


def test_convert_the_banana_no_name(open_converted=False):

    pfo_study_in = os.path.join(root_dir, 'test_data', 'bru_banana')
    pfo_study_out = os.path.join(root_dir, 'test_data', 'nifti_banana')

    # delete study if already exists:
    target_folder = os.path.join(pfo_study_out, 'APMFruits20111130')
    if os.path.exists(target_folder):
        os.system('rm -r {}'.format(os.path.join(target_folder)))

    bru = Bruker2Nifti(pfo_study_in, pfo_study_out)

    bru.verbose       = 2,
    bru.correct_slope = True,
    bru.get_acqp      = False,
    bru.get_method    = False,
    bru.get_reco      = False
    bru.convert()

    if open_converted:

        if platform.system() == "Windows":
            os.startfile(pfo_study_out.encode('string-escape'))
        elif platform.system() == "Darwin":
            subprocess.Popen(["open", pfo_study_out])
        else:
            subprocess.Popen(["xdg-open", pfo_study_out])

    for ex in ['1', '2', '3']:
        experiment_folder = os.path.join(pfo_study_out, 'APMFruits20111130', 'APMFruits20111130_{}'.format(ex))
        assert os.path.exists(experiment_folder)
        assert os.path.exists(os.path.join(experiment_folder, 'APMFruits20111130_{}.nii.gz'.format(ex)))


def test_warning_banana_bad_n():

    for n in ['1', '2', '3']:

        pfo_study_in = os.path.join(root_dir, 'test_data', 'bru_banana_bad_' + n)
        pfo_study_out = os.path.join(root_dir, 'test_data', 'nifti_banana')

        with warnings.catch_warnings(record=True) as wa:

            warnings.simplefilter('always')

            bru = Bruker2Nifti(pfo_study_in, pfo_study_out, study_name='banana')
            bru.correct_slope = True
            bru.verbose = 2
            bru.convert()

            assert len(wa) > 0
            assert issubclass(wa[-1].category, Warning)



#!/usr/bin/env python
import os
from setuptools import setup, find_packages

from bruker2nifti.__init__ import __version__ as version


def requirements2list(pfi_txt='requirements.txt'):
    here = os.path.dirname(os.path.realpath(__file__))
    f = open(os.path.join(here, pfi_txt), 'r')
    l = []
    for line in f.readlines():
        l.append(line.replace('\n', ''))
    return l


infos = {
         'name': 'bruker2nifti',
         'version': '1.0.3',
         'description': 'From raw Bruker to nifti, home-made converter.',
         'web_infos' : '',
         'repository': {
                        'type': 'git',
                        'url': 'https://github.com/SebastianoF/bruker2nifti'
                       },
         'author': 'sebastiano ferraris',
         'author_email': 'sebastiano.ferraris@gmail.com',
         'dependencies': {
                          # requirements.txt file automatically generated using pipreqs.
                          'python' : 'requirements.txt'
                          }
         }

setup(name=infos['name'],
      version=infos['version'],
      description=infos['description'],
      author=infos['author'],
      author_email=infos['author_email'],
      url=infos['repository']['url'],
      packages=find_packages(),
      install_requires=requirements2list(),
      entry_points={
        'console_scripts': ['bruker2nifti=bruker2nifti.cli.bruker2nii:main',
                            'bruker2nifti_scan=bruker2nifti.cli.bruker2nii_scan:main_scan'],
        'gui_scripts' : ['bruker2nifti_gui=bruker2nifti.gui.open:open_gui']
      }
      )



import os
import nibabel as nib
import sys
import numpy as np
import warnings

from os.path import join as jph

from ._getters import get_list_scans, nifti_getter
from ._utils import bruker_read_files, normalise_b_vect, from_dict_to_txt_sorted, set_new_data, \
    apply_reorientation_to_b_vects, obtain_b_vectors_orient_matrix


def scan2struct(pfo_scan,
                correct_slope=True,
                correct_offset=True,
                sample_upside_down=False,
                nifti_version=1,
                qform_code=1,
                sform_code=2,
                get_acqp=False,
                get_method=False,
                get_reco=False,
                frame_body_as_frame_head=False,
                keep_same_det=True,
                consider_subject_position=False
                ):
    """
    The core method of the converter has 2 parts.
    1) parsing the Bruker scan folder structure into an internal dictionary called struct.
    2) writing the information parsed in struct into folders.
    ----
    scan2struct is the first part of the bridge. Info required to fill nifti header are in the visu_pars file.
    The user may want to parse as well acqp, method (must when EpiDti) and reco parameter files.
    Data are parsed in the intermediate dictionary struct containing the final scan(s) converted in nibabel
    image, with additional infos.
    :param pfo_scan: path to folder containing the scan
    :param correct_slope: [True] if you want to correct the slope of the values.
    :param correct_offset: [True] if you want to correct the offset of the values.
    :param sample_upside_down: [False] if you want to have the sample rotated 180 around the Anterior-Posterior axis.
    :param nifti_version: [1] output nifti version can be version 1 or version 2 (see nibabel documentation)
    :param qform_code: [1] qform of the final nifti image
    :param sform_code: [2] sform of the final nifti image
    :param get_acqp: [False] if you want to parse the information in the acqp parameter file of the bruker raw data
    :param get_method: [False] if you want to parse the information in the method file. Forced to True when
    dealing with diffusion weighted images.
    :param get_reco: [False] if you want to parse the information in the reco parameter file.
    :param frame_body_as_frame_head: e.g. true if monkey, false if rat.
    :param keep_same_det: impose to have in the nifti affine matrix, the same determinat as in the bruker parameter.
    :param consider_subject_position : visu_pars SubjPosition can be 'Head_Prone' or 'Head_Supine'. While it may
    make sense in most cases to take this value into account, in some other it may not, as it is
    tuned to switch from radiological to neurological coordinate systems in a work-around.
    If the subject is Prone and the technician wants to have the coordinates
    in neurological he/she can consciously set the variable vc_subject_position to 'Head_Supine'.
    :return: output_data data structure containing the nibabel image(s) {nib_list, visu_pars_list, acqp, method, reco}
    """

    if not os.path.isdir(pfo_scan):
        raise IOError('Input folder does not exists.')

    # Get system endian_nes
    system_endian_nes = sys.byteorder

    # Get sub-scans series in the same experiment.
    list_sub_scans = get_list_scans(jph(pfo_scan, 'pdata'))

    if not list_sub_scans:
        warn_msg = "\nNo sub scan in the folder structure: \n{}. \nAre you sure the input folder contains a " \
                   "proper Bruker scan?\n".format(jph(pfo_scan, 'pdata'))
        warnings.warn(warn_msg)
        return None

    nib_scans_list = []
    visu_pars_list = []

    for id_sub_scan in list_sub_scans:

        visu_pars = bruker_read_files('visu_pars', pfo_scan, sub_scan_num=id_sub_scan)

        if visu_pars == {}:
            warn_msg = "\nNo 'visu_pars' data found here: \n{}. \nAre you sure the input folder contains a " \
                       "proper Bruker scan?\n".format(jph(pfo_scan, 'pdata', id_sub_scan))
            warnings.warn(warn_msg)
            return None

        # In some cases we cannot deal with, VisuPars['VisuCoreSize'] can be a float. No conversion in this case.
        if not (isinstance(visu_pars['VisuCoreSize'], np.ndarray) or isinstance(visu_pars['VisuCoreSize'], list)):
            warnings.warn("\nWarning, VisuCoreSize in VisuPars parameter file {} \n"
                          "is not a list or a vector in. The study cannot be converted. \n"
                          "".format(jph(pfo_scan, 'pdata', id_sub_scan)))
            return None

        # Get data endian_nes - default big!!
        if visu_pars['VisuCoreByteOrder'] == 'littleEndian':
            data_endian_ness = 'little'
        elif visu_pars['VisuCoreByteOrder'] == 'bigEndian':
            data_endian_ness = 'big'
        else:
            data_endian_ness = 'big'

        # Get datatype
        if visu_pars['VisuCoreWordType'] == '_32BIT_SGN_INT':
            dt = np.int32
        elif visu_pars['VisuCoreWordType'] == '_16BIT_SGN_INT':
            dt = np.int16
        elif visu_pars['VisuCoreWordType'] == '_8BIT_UNSGN_INT':
            dt = np.uint8
        elif visu_pars['VisuCoreWordType'] == '_32BIT_FLOAT':
            dt = np.float32
        else:
            raise IOError('Unknown data type for VisuPars VisuCoreWordType')

        # GET IMAGE VOLUME
        if os.path.exists(jph(pfo_scan, 'pdata', id_sub_scan, '2dseq')):
            img_data_vol = np.copy(np.fromfile(jph(pfo_scan, 'pdata', id_sub_scan, '2dseq'), dtype=dt))
        else:
            warn_msg = "\nNo '2dseq' data found here: \n{}. \nAre you sure the input folder contains a " \
                       "proper Bruker scan?\n".format(jph(pfo_scan, 'pdata', id_sub_scan))
            warnings.warn(warn_msg)
            return None

        if not data_endian_ness == system_endian_nes:
            img_data_vol.byteswap(True)

        if 'VisuAcqSequenceName' in visu_pars.keys():
            visu_pars_acq_sequence_name = visu_pars['VisuAcqSequenceName']
        else:
            visu_pars_acq_sequence_name = ''

        is_dwi = 'dtiepi' in visu_pars_acq_sequence_name.lower()

        if is_dwi:
            # Force to not correcting the slope, if true. Diffusion weighted images must be slope corrected before the
            # DTI analysis. They will be to heavy otherwise.
            correct_slope = False
            correct_offset = False
            # Force method to be parsed. Useful infos in this file to process the DWI.
            get_method = True

        # ------------------------------------------------------ #
        # ------ Generate the nifti image using visu_pars. ----- #
        # ------------------------------------------------------ #

        print('sample upside down before nib {}'.format(sample_upside_down))
        nib_im = nifti_getter(img_data_vol,
                              visu_pars,
                              correct_slope,
                              correct_offset,
                              sample_upside_down,
                              nifti_version,
                              qform_code,
                              sform_code,
                              frame_body_as_frame_head=frame_body_as_frame_head,
                              keep_same_det=keep_same_det,
                              consider_subject_position=consider_subject_position
                              )
        # ------------------------------------------------------ #
        # ------------------------------------------------------ #

        nib_scans_list.append(nib_im)
        visu_pars_list.append(visu_pars)

    # -- Get additional data

    # Get information from method, if it exists. Parse Method parameter and erase the dictionary if unwanted
    method = bruker_read_files('method', pfo_scan)

    if method == {}:
        print("Warning: No 'method' file to parse.")
    if 'Method' in method.keys():
        acquisition_method = method['Method'].replace('<', '').replace('>', '').split(':')[-1]
    else:
        acquisition_method = ''

    if not get_method:
        method = {}

    # Get information from acqp, reco, if they exist.
    acqp = {}
    reco = {}

    if get_acqp:
        acqp = bruker_read_files('acqp', pfo_scan)
        if acqp == {}:
            print("Warning: No 'acqp' file to parse.")

    if get_reco:
        reco = bruker_read_files('reco', pfo_scan)
        if reco == {}:
            print("Warning: No 'method' file to parse.")

    # -- Return data structure
    struct_scan = {'nib_scans_list' : nib_scans_list,
                   'visu_pars_list' : visu_pars_list,
                   'acqp'           : acqp,
                   'reco'           : reco,
                   'method'         : method,
                   'acquisition_method'    : acquisition_method}

    return struct_scan


def write_struct(bruker_struct,
                 pfo_output,
                 fin_scan='',
                 save_human_readable=True,
                 save_b0_if_dwi=True,
                 verbose=1,
                 frame_body_as_frame_head=False,
                 keep_same_det=True,
                 consider_subject_position=False,
                 ):
    """
    The core method of the converter has 2 parts.
    1) parsing the Bruker scan folder structure into an internal dictionary called struct.
    2) writing the information parsed in struct into folders.
    -------
    write_struct is the second part of the bridge -
    :param bruker_struct: output of scan2struct
    :param pfo_output: path-to-folder where the converted structure will be saved.
    :param fin_scan: filename of the scan
    :param save_human_readable: output data will be saved in .txt other than in numpy format.
    :param save_b0_if_dwi: save the first time-point if the data is a DWI.
    :param verbose:
    :param frame_body_as_frame_head: according to the animal. If True monkey, if False rat-rabbit
    :param keep_same_det: force the initial determinant to be the same as the final one
    :param consider_subject_position: Attribute manually set, or left blank, by the lab experts. False by default
    :return: save the bruker_struct parsed in scan2struct in the specified folder, with the specified parameters.
    """

    if not os.path.isdir(pfo_output):
        raise IOError('Output folder does not exist.')

    if bruker_struct is None:
        return

    if not len(bruker_struct['visu_pars_list']) == len(bruker_struct['nib_scans_list']):
        raise IOError('Visu pars list and scans list have a different number of elements.')

    if fin_scan is None:
        fin_scan = ''

    # -- WRITE Additional data shared by all the sub-scans:
    # if the modality is a DtiEpi or Dwimage then save the DW directions, b values and b vectors in separate csv .txt.

    is_dwi = 'dtiepi' in bruker_struct['visu_pars_list'][0]['VisuAcqSequenceName'].lower() or \
                  'dwi' in bruker_struct['visu_pars_list'][0]['VisuAcqSequenceName'].lower()

    if is_dwi:  # File method is the same for each sub-scan. Cannot embed this in the next for cycle.

        # -- Deals with b-vector: normalise, reorient and save in external .npy/txt.
        dw_grad_vec = bruker_struct['method']['DwGradVec']

        assert dw_grad_vec.shape[0] == bruker_struct['method']['DwNDiffExp']

        # get b-vectors re-orientation matrix from visu-pars
        reorientation_matrix = obtain_b_vectors_orient_matrix(bruker_struct['visu_pars_list'][0]['VisuCoreOrientation'],
                                                              bruker_struct['visu_pars_list'][0]['VisuSubjectPosition'],
                                                              frame_body_as_frame_head=frame_body_as_frame_head,
                                                              keep_same_det=keep_same_det,
                                                              consider_subject_position=consider_subject_position)

        # apply reorientation
        dw_grad_vec = apply_reorientation_to_b_vects(reorientation_matrix, dw_grad_vec)
        # normalise:
        dw_grad_vec = normalise_b_vect(dw_grad_vec)

        np.save(jph(pfo_output, fin_scan + '_DwGradVec.npy'), dw_grad_vec)

        if save_human_readable:
            np.savetxt(jph(pfo_output, fin_scan + '_DwGradVec.txt'), dw_grad_vec, fmt='%.14f')

        if verbose > 0:
            msg = 'Diffusion weighted directions saved in ' + jph(pfo_output, fin_scan + '_DwDir.npy')
            print(msg)

        b_vals = bruker_struct['method']['DwEffBval']
        b_vects = bruker_struct['method']['DwDir']

        np.save(jph(pfo_output, fin_scan + '_DwEffBval.npy'), b_vals)
        np.save(jph(pfo_output, fin_scan + '_DwDir.npy'), b_vects)

        if save_human_readable:
            np.savetxt(jph(pfo_output, fin_scan + '_DwEffBval.txt'), b_vals, fmt='%.14f')
            np.savetxt(jph(pfo_output, fin_scan + '_DwDir.txt'), b_vects, fmt='%.14f')

        if verbose > 0:
            print('B-vectors saved in {}'.format(jph(pfo_output, fin_scan + '_DwEffBval.npy')))
            print('B-values  saved in {}'.format(jph(pfo_output, fin_scan + '_DwGradVec.npy')))

    # save the dictionary as numpy array containing the corresponding dictionaries
    # TODO use pickle instead of numpy to save the dictionaries(?)

    if not bruker_struct['acqp'] == {}:
        np.save(jph(pfo_output, fin_scan + '_acqp.npy'), bruker_struct['acqp'])
        if save_human_readable:
            from_dict_to_txt_sorted(bruker_struct['acqp'], jph(pfo_output, fin_scan + '_acqp.txt'))
    if not bruker_struct['method'] == {}:
        np.save(jph(pfo_output, fin_scan + '_method.npy'), bruker_struct['method'])
        if save_human_readable:
            from_dict_to_txt_sorted(bruker_struct['method'], jph(pfo_output, fin_scan + '_method.txt'))
    if not bruker_struct['reco'] == {}:
        np.save(jph(pfo_output, fin_scan + '_reco.npy'), bruker_struct['reco'])
        if save_human_readable:
            from_dict_to_txt_sorted(bruker_struct['reco'], jph(pfo_output, fin_scan + '_reco.txt'))

    # Visu_pars and summary info for each sub-scan:
    summary_info = {}

    for i in range(len(bruker_struct['visu_pars_list'])):

        if len(bruker_struct['nib_scans_list']) > 1:
            i_label = '_subscan_' + str(i) + '_'
        else:
            i_label = '_'

        # A) Save visu_pars for each sub-scan:
        np.save(jph(pfo_output, fin_scan + i_label + 'visu_pars.npy'), bruker_struct['visu_pars_list'][i])

        # B) Save single slope data for each sub-scan (from visu_pars):
        np.save(jph(pfo_output, fin_scan + i_label + 'slope.npy'),
                bruker_struct['visu_pars_list'][i]['VisuCoreDataSlope'])

        # A and B) save them both in .txt if human readable version of data is required.
        if save_human_readable:
            from_dict_to_txt_sorted(bruker_struct['visu_pars_list'][i],
                                    jph(pfo_output, fin_scan + i_label + 'visu_pars.txt'))

            slope = bruker_struct['visu_pars_list'][i]['VisuCoreDataSlope']
            if not isinstance(slope, np.ndarray):
                slope = np.atleast_2d(slope)
            np.savetxt(jph(pfo_output, fin_scan + i_label + 'slope.txt'), slope, fmt='%.14f')

        # Update summary dictionary:
        summary_info_i = {i_label[1:] + "visu_pars['VisuUid']"
                                    : bruker_struct['visu_pars_list'][i]['VisuUid'],
                          i_label[1:] + "visu_pars['VisuCoreDataSlope']"
                                    : bruker_struct['visu_pars_list'][i]['VisuCoreDataSlope'],
                          i_label[1:] + "visu_pars['VisuCoreSize']"
                                    : bruker_struct['visu_pars_list'][i]['VisuCoreSize'],
                          i_label[1:] + "visu_pars['VisuCoreOrientation']"
                                    : bruker_struct['visu_pars_list'][i]['VisuCoreOrientation'],
                          i_label[1:] + "visu_pars['VisuCorePosition']"
                                    : bruker_struct['visu_pars_list'][i]['VisuCorePosition']}

        if len(list(bruker_struct['visu_pars_list'][i]['VisuCoreExtent'])) == 2:
            # equivalent to struct['method']['SpatDimEnum'] == '2D':
            if 'VisuCoreSlicePacksSlices' in bruker_struct['visu_pars_list'][i].keys():
                summary_info_i.update({i_label[1:] + "visu_pars['VisuCoreSlicePacksSlices']"
                                                : bruker_struct['visu_pars_list'][i]['VisuCoreSlicePacksSlices']})

        if len(list(bruker_struct['visu_pars_list'][i]['VisuCoreExtent'])) == 3 and \
                        'VisuCoreDiskSliceOrder' in bruker_struct['visu_pars_list'][i].keys():
            # first part equivalent to struct['method']['SpatDimEnum'] == '3D':
            summary_info_i.update({i_label[1:] + "visu_pars['VisuCoreDiskSliceOrder']"
                                            : bruker_struct['visu_pars_list'][i]['VisuCoreDiskSliceOrder']})

        if 'VisuCreatorVersion' in bruker_struct['visu_pars_list'][i].keys():
            summary_info_i.update({i_label[1:] + "visu_pars['VisuCreatorVersion']"
                                            : bruker_struct['visu_pars_list'][i]['VisuCreatorVersion']})

        summary_info.update(summary_info_i)

        # WRITE NIFTI IMAGES:

        if isinstance(bruker_struct['nib_scans_list'][i], list):
            # the scan had sub-volumes embedded. they are saved separately
            for sub_vol_id, subvol in enumerate(bruker_struct['nib_scans_list'][i]):

                if fin_scan == '':
                    pfi_scan = jph(pfo_output, 'scan' + i_label[:-1] + '_subvol_' + str(sub_vol_id) + '.nii.gz')
                else:
                    pfi_scan = jph(pfo_output, fin_scan + i_label[:-1] + '_subvol_' + str(sub_vol_id) + '.nii.gz')

                nib.save(subvol, pfi_scan)

        else:

            if fin_scan == '':
                pfi_scan = jph(pfo_output, 'scan' + i_label[:-1] + '.nii.gz')
            else:
                pfi_scan = jph(pfo_output, fin_scan + i_label[:-1] + '.nii.gz')

            nib.save(bruker_struct['nib_scans_list'][i], pfi_scan)

            if save_b0_if_dwi and is_dwi:
                # save the b0, first slice alone. Optimized if you have
                # NiftiSeg (http://cmictig.cs.ucl.ac.uk/wiki/index.php/NiftySeg) installed

                if fin_scan == '':
                    pfi_scan_b0 = jph(pfo_output, 'scan' + i_label[:-1] + '_b0.nii.gz')
                else:
                    pfi_scan_b0 = jph(pfo_output, fin_scan + i_label[:-1] + '_b0.nii.gz')

                nib.save(set_new_data(bruker_struct['nib_scans_list'][i],
                                      bruker_struct['nib_scans_list'][i].get_data()[..., 0]),
                         pfi_scan_b0)
                if verbose > 0:
                    msg = 'b0 scan saved alone in ' + pfi_scan_b0
                    print(msg)

    # complete the summary info with additional information from other parameter files, if required:

    if not bruker_struct['acqp'] == {}:

        summary_info_acqp = {"acqp['ACQ_sw_version']"     : bruker_struct['acqp']['ACQ_sw_version'],
                             "acqp['NR']"                 : bruker_struct['acqp']['NR'],
                             "acqp['NI']"                 : bruker_struct['acqp']['NI'],
                             "acqp['ACQ_n_echo_images']"  : bruker_struct['acqp']['ACQ_n_echo_images'],
                             "acqp['ACQ_slice_thick']"    : bruker_struct['acqp']['ACQ_slice_thick']}
        summary_info.update(summary_info_acqp)

    if not bruker_struct['method'] == {}:

        summary_info_method = {"method['SpatDimEnum']"         : bruker_struct['method']['SpatDimEnum'],
                               "method['Matrix']"              : bruker_struct['method']['Matrix'],
                               "method['SpatResol']"           : bruker_struct['method']['SpatResol'],
                               "method['Method']"              : bruker_struct['method']['Method'],
                               "method['SPackArrSliceOrient']" : bruker_struct['method']['SPackArrSliceOrient'],
                               "method['SPackArrReadOrient']"  : bruker_struct['method']['SPackArrReadOrient']}
        summary_info.update(summary_info_method)

    if not bruker_struct['reco'] == {}:

        summary_info_reco = {"reco['RECO_size']"       : bruker_struct['reco']['RECO_size'],
                             "reco['RECO_inp_order']"  : bruker_struct['reco']['RECO_inp_order']}
        summary_info.update(summary_info_reco)

    # Finally summary info with the updated information.
    from_dict_to_txt_sorted(summary_info, jph(pfo_output, fin_scan + '_summary.txt'))

    # Get the method name in a single .txt file:
    if bruker_struct['acquisition_method'] is not '':
        text_file = open(jph(pfo_output, 'acquisition_method.txt'), "w+")
        text_file.write(bruker_struct['acquisition_method'])
        text_file.close()



from subprocess import check_output
import os


# # Describe the version relative to last tag
# command_git = ['git', 'describe', '--match', 'v[0-9]*']
# version_buf = check_output(command_git).rstrip()
#
# print(version_buf)
#
# # Exclude the 'v' for PEP440 conformity, see
# # https://www.python.org/dev/peps/pep-0440/#public-version-identifiers
# version_buf = version_buf[1:]
#
# print(version_buf)

import os

from bruker2nifti.converter import Bruker2Nifti


if __name__ == '__main__':
    pfo_study_in = os.path.join('/Users/sebastiano/Desktop/test_converter/1305')
    pfo_study_out = '/Users/sebastiano/Desktop/test_converter/converted'
    print(os.path.isdir(pfo_study_out))
    # instantiate a converter
    bru = Bruker2Nifti(pfo_study_in, pfo_study_out, study_name='1305NoOffset')
    # select the options (attributes) you may want to change - the one shown below are the default one:
    bru.verbose = 2
    bru.correct_slope = True
    bru.correct_offset = False
    bru.get_acqp = True
    bru.get_method = True
    bru.get_reco = True
    bru.nifti_version = 1
    bru.qform_code = 1
    bru.sform_code = 2
    bru.save_human_readable = True
    bru.save_b0_if_dwi = True

    # Check that the list of scans and the scans names automatically selected makes some sense:
    print(bru.scans_list)
    print(bru.list_new_name_each_scan)
    # call the function convert, to convert the study:
    bru.convert()

